from ..abstract_discrete_distribution import AbstractLDDiscreteDistribution
from ...util import ParameterError, ParameterWarning
import qmctoolscl
from os.path import dirname, abspath, isfile
import numpy as np
from numpy.lib.npyio import DataSource
import warnings
from copy import deepcopy


class DigitalNetB2(AbstractLDDiscreteDistribution):
    r"""
    Low discrepancy digital net in base 2.

    Note:
        - Digital net sample sizes should be powers of $2$ e.g. $1$, $2$, $4$, $8$, $16$, $\dots$.
        - The first point of an unrandomized digital nets is the origin.
        - `Sobol` is an alias for `DigitalNetB2`.
        - To use higher order digital nets, either:
            
            - Pass in `generating_matrices` *without* interlacing and supply `alpha`>1 to apply interlacing, or 
            - Pass in `generating_matrices` *with* interlacing and set `alpha=1` to avoid additional interlacing  
            
            i.e. do *not* pass in interlaced `generating_matrices` and set `alpha>1`, this will apply additional interlacing. 
    
    Examples:
        >>> discrete_distrib = DigitalNetB2(2,seed=7)
        >>> discrete_distrib(4)
        array([[0.72162356, 0.914955  ],
               [0.16345554, 0.42964856],
               [0.98676255, 0.03436384],
               [0.42956655, 0.55876342]])
        >>> discrete_distrib(1) # first point in the sequence
        array([[0.72162356, 0.914955  ]])
        >>> discrete_distrib
        DigitalNetB2 (AbstractLDDiscreteDistribution)
            d               2^(1)
            replications    1
            randomize       LMS DS
            gen_mats_source joe_kuo.6.21201.txt
            order           RADICAL INVERSE
            t               63
            alpha           1
            n_limit         2^(32)
            entropy         7

        Replications of independent randomizations

        >>> x = DigitalNetB2(dimension=3,seed=7,replications=2)(4)
        >>> x.shape
        (2, 4, 3)
        >>> x
        array([[[0.24653277, 0.1821862 , 0.74732591],
                [0.68152903, 0.66169442, 0.42891961],
                [0.48139855, 0.79818233, 0.08201287],
                [0.91541325, 0.29520621, 0.77495809]],
        <BLANKLINE>
               [[0.44876891, 0.85899604, 0.50549679],
                [0.53635924, 0.04353443, 0.33564946],
                [0.23214143, 0.29281506, 0.06841036],
                [0.75295715, 0.60241448, 0.76962976]]])

        Different orderings (avoid warnings that the first point is the origin)

        >>> DigitalNetB2(dimension=2,randomize=False,order="GRAY")(n_min=2,n_max=4,warn=False)
        array([[0.75, 0.25],
               [0.25, 0.75]])
        >>> DigitalNetB2(dimension=2,randomize=False,order="RADICAL INVERSE")(n_min=2,n_max=4,warn=False)
        array([[0.25, 0.75],
               [0.75, 0.25]])
        
        Generating matrices from [https://github.com/QMCSoftware/LDData/tree/main/dnet](https://github.com/QMCSoftware/LDData/tree/main/dnet)

        >>> DigitalNetB2(dimension=3,randomize=False,generating_matrices="mps.nx_s5_alpha2_m32.txt")(8,warn=False)
        array([[0.        , 0.        , 0.        ],
               [0.75841841, 0.45284834, 0.48844557],
               [0.57679828, 0.13226272, 0.10061957],
               [0.31858402, 0.32113875, 0.39369111],
               [0.90278927, 0.45867532, 0.01803333],
               [0.14542431, 0.02548793, 0.4749614 ],
               [0.45587539, 0.33081476, 0.11474426],
               [0.71318879, 0.15377192, 0.37629925]])
        
        All randomizations 

        >>> DigitalNetB2(dimension=3,randomize='LMS DS',seed=5)(8)
        array([[0.69346401, 0.20118185, 0.64779396],
               [0.43998032, 0.90102467, 0.0936172 ],
               [0.86663563, 0.60910036, 0.26043276],
               [0.11327376, 0.30772653, 0.93959283],
               [0.62102883, 0.79169756, 0.77051637],
               [0.37451038, 0.1231324 , 0.46634012],
               [0.94785596, 0.38577413, 0.13377215],
               [0.20121617, 0.71843325, 0.56293458]])
        >>> DigitalNetB2(dimension=3,randomize='LMS',seed=5)(8,warn=False)
        array([[0.        , 0.        , 0.        ],
               [0.75446077, 0.83265937, 0.69584079],
               [0.42329494, 0.65793842, 0.90427279],
               [0.67763292, 0.48937304, 0.33344964],
               [0.18550714, 0.97332905, 0.3772791 ],
               [0.93104851, 0.17195496, 0.82311652],
               [0.26221346, 0.31742386, 0.53093284],
               [0.50787715, 0.5172669 , 0.2101083 ]])
        >>> DigitalNetB2(dimension=3,randomize='DS',seed=5)(8)
        array([[0.68383949, 0.04047995, 0.42903182],
               [0.18383949, 0.54047995, 0.92903182],
               [0.93383949, 0.79047995, 0.67903182],
               [0.43383949, 0.29047995, 0.17903182],
               [0.55883949, 0.66547995, 0.05403182],
               [0.05883949, 0.16547995, 0.55403182],
               [0.80883949, 0.41547995, 0.80403182],
               [0.30883949, 0.91547995, 0.30403182]])
        >>> DigitalNetB2(dimension=3,randomize='OWEN',seed=5)(8)
        array([[0.33595486, 0.05834975, 0.30066401],
               [0.89110875, 0.84905188, 0.81833285],
               [0.06846074, 0.59997956, 0.67064205],
               [0.6693703 , 0.25824002, 0.10469644],
               [0.44586618, 0.99161977, 0.1873488 ],
               [0.84245267, 0.16445553, 0.56544372],
               [0.18546359, 0.44859876, 0.97389524],
               [0.61215442, 0.64341386, 0.44529863]])
        
        Higher order net without randomization 
        
        >>> DigitalNetB2(dimension=3,randomize='FALSE',seed=7,alpha=2)(4,warn=False)
        array([[0.    , 0.    , 0.    ],
               [0.75  , 0.75  , 0.75  ],
               [0.4375, 0.9375, 0.1875],
               [0.6875, 0.1875, 0.9375]])


        Higher order nets with randomizations and replications 
                       
        >>> DigitalNetB2(dimension=3,randomize='LMS DS',seed=7,replications=2,alpha=2)(4,warn=False)
        array([[[0.42955149, 0.89149058, 0.43867111],
                [0.68701828, 0.07601148, 0.51312447],
                [0.10088033, 0.16293661, 0.25144138],
                [0.85846252, 0.87103178, 0.70041789]],
        <BLANKLINE>
               [[0.27151905, 0.42406763, 0.21917369],
                [0.55035224, 0.67864387, 0.90033876],
                [0.19356758, 0.57589964, 0.00347701],
                [0.97235125, 0.32168581, 0.86920948]]])
        >>> DigitalNetB2(dimension=3,randomize='LMS',seed=7,replications=2,alpha=2)(4,warn=False)
        array([[[0.        , 0.        , 0.        ],
                [0.75817062, 0.96603053, 0.94947625],
                [0.45367986, 0.80295638, 0.18778553],
                [0.71171791, 0.2295424 , 0.76175441]],
        <BLANKLINE>
               [[0.        , 0.        , 0.        ],
                [0.78664636, 0.75470215, 0.86876474],
                [0.45336727, 0.99953621, 0.22253579],
                [0.73996397, 0.24544824, 0.9008679 ]]])
        >>> DigitalNetB2(dimension=3,randomize='DS',seed=7,replications=2,alpha=2)(4)
        array([[[0.04386058, 0.58727432, 0.3691824 ],
                [0.79386058, 0.33727432, 0.6191824 ],
                [0.48136058, 0.39977432, 0.4316824 ],
                [0.73136058, 0.64977432, 0.6816824 ]],
        <BLANKLINE>
               [[0.65212985, 0.69669968, 0.10605352],
                [0.40212985, 0.44669968, 0.85605352],
                [0.83962985, 0.25919968, 0.16855352],
                [0.08962985, 0.50919968, 0.91855352]]])
        >>> DigitalNetB2(dimension=3,randomize='OWEN',seed=7,replications=2,alpha=2)(4)
        array([[[0.46368517, 0.03964427, 0.62172094],
                [0.7498683 , 0.76141348, 0.4243043 ],
                [0.01729754, 0.97968459, 0.65963223],
                [0.75365329, 0.1903774 , 0.34141493]],
        <BLANKLINE>
               [[0.52252547, 0.5679709 , 0.05949112],
                [0.27248656, 0.36488289, 0.81844058],
                [0.94219959, 0.39172304, 0.20285965],
                [0.19716391, 0.64741585, 0.92494554]]])
    
    **References:**

    1.  Marius Hofert and Christiane Lemieux.  
        qrng: (Randomized) Quasi-Random Number Generators (2019).  
        R package version 0.0-7.  
        [https://CRAN.R-project.org/package=qrng](https://CRAN.R-project.org/package=qrng).

    2.  Faure, Henri, and Christiane Lemieux.  
        Implementation of Irreducible Sobol' Sequences in Prime Power Bases.  
        Mathematics and Computers in Simulation 161 (2019): 13-22. Crossref. Web.

    3.  F.Y. Kuo, D. Nuyens.  
        Application of quasi-Monte Carlo methods to elliptic PDEs with random diffusion coefficients \- a survey of analysis and implementation.  
        Foundations of Computational Mathematics, 16(6):1631-1696, 2016.  
        [https://link.springer.com/article/10.1007/s10208-016-9329-5](https://link.springer.com/article/10.1007/s10208-016-9329-5). 
        
    4.  D. Nuyens.  
        The Magic Point Shop of QMC point generators and generating vectors.  
        MATLAB and Python software, 2018.  
        [https://people.cs.kuleuven.be/~dirk.nuyens/](https://people.cs.kuleuven.be/~dirk.nuyens/).

    5.  R. Cools, F.Y. Kuo, D. Nuyens.  
        Constructing embedded lattice rules for multivariate integration.  
        SIAM J. Sci. Comput., 28(6), 2162-2188.

    6.  I.M. Sobol', V.I. Turchaninov, Yu.L. Levitan, B.V. Shukhman.  
        Quasi-Random Sequence Generators.  
        Keldysh Institute of Applied Mathematics.  
        Russian Academy of Sciences, Moscow (1992).

    7.  Sobol, Ilya & Asotsky, Danil & Kreinin, Alexander & Kucherenko, Sergei. (2011).  
        Construction and Comparison of High-Dimensional Sobol' Generators. Wilmott. 2011.  
        [10.1002/wilm.10056](https://onlinelibrary.wiley.com/doi/abs/10.1002/wilm.10056).

    8.  Paul Bratley and Bennett L. Fox.  
        Algorithm 659: Implementing Sobol's quasirandom sequence generator.  
        ACM Trans. Math. Softw. 14, 1 (March 1988), 88-100. 1988.  
        [https://doi.org/10.1145/42288.214372](https://doi.org/10.1145/42288.2143720).
    """

    def __init__(self,
                 dimension = 1,
                 replications = None,
                 seed = None,
                 randomize = 'LMS DS',
                 generating_matrices = "joe_kuo.6.21201.txt",
                 order = 'RADICAL INVERSE',
                 t = 63,
                 alpha = 1,
                 msb = None,
                 _verbose = False,
                 # deprecated
                 graycode = None,
                 t_max = None,
                 t_lms = None):
        r"""
        Args:
            dimension (Union[int,np.ndarray]): Dimension of the generator.

                - If an `int` is passed in, use generating vector components at indices 0,...,`dimension`-1.
                - If an `np.ndarray` is passed in, use generating vector components at these indices.
            
            replications (int): Number of independent randomizations of a pointset.
            seed (Union[None,int,np.random.SeedSeq): Seed the random number generator for reproducibility.
            randomize (str): Options are
                
                - `'LMS DS'`: Linear matrix scramble with digital shift.
                - `'LMS'`: Linear matrix scramble only.
                - `'DS'`: Digital shift only.
                - `'NUS'`: Nested uniform scrambling. Also known as Owen scrambling. 
                - `'FALSE'`: No randomization. In this case the first point will be the origin. 

            generating_matrices (Union[str,np.ndarray,int]: Specify the generating matrices.
                
                - A `str` should be the name (or path) of a file from the LDData repo at [https://github.com/QMCSoftware/LDData/tree/main/dnet](https://github.com/QMCSoftware/LDData/tree/main/dnet).
                - An `np.ndarray` of integers with shape $(d,m_\mathrm{max})$ or $(r,d,m_\mathrm{max})$ where $d$ is the number of dimensions, $r$ is the number of replications, and $2^{m_\mathrm{max}}$ is the maximum number of supported points. Setting `msb=False` will flip the bits of ints in the generating matrices.
            
            order (str): `'RADICAL INVERSE'`, or `'GRAY'` ordering. See the doctest example above.
            t (int): Number of bits in integer represetation of points *after* randomization. The number of bits in the generating matrices is inferred based on the largest value.
            alpha (int): Interlacing factor for higher order nets.  
                When `alpha`>1, interlacing is performed regardless of the generating matrices,  
                i.e., for `alpha`>1 do *not* pass in generating matrices which are already interlaced.  
                The Note for this class contains more info.  
            msb (bool): Flag for Most Significant Bit (MSB) vs Least Significant Bit (LSB) integer representations in generating matrices. If `msb=False` (LSB order), then integers in generating matrices will be bit-reversed. 
            _verbose (bool): If `True`, print linear matrix scrambling matrices. 
        """
        if graycode is not None:
            order = 'GRAY' if graycode else 'RADICAL INVERSE'
            warnings.warn("graycode argument deprecated, set order='GRAY' or order='RADICAL INVERSE' instead. Using order='%s'"%order,ParameterWarning)
        if t_lms is not None:
            t = t_lms
            warnings.warn("t_lms argument deprecated. Set t instead. Using t = %d"%t,ParameterWarning)
        if t_max is not None: 
            warnings.warn("t_max is deprecated as it can be inferred from the generating matrices. Set t to change the number of bits after randomization.",ParameterWarning)
        self.parameters = ['randomize','gen_mats_source','order','t','alpha','n_limit']
        self.input_generating_matrices = deepcopy(generating_matrices)
        self.input_t = deepcopy(t) 
        self.input_msb = deepcopy(msb)
        if isinstance(generating_matrices,str) and generating_matrices=="joe_kuo.6.21201.txt":
            self.gen_mats_source = generating_matrices
            if np.isscalar(dimension) and dimension<=1024 and alpha==1:
                gen_mats = np.load(dirname(abspath(__file__))+'/generating_matrices/joe_kuo.6.1024.npy')[None,:]
                d_limit = 1024
            else:
                gen_mats = np.load(dirname(abspath(__file__))+'/generating_matrices/joe_kuo.6.21201.npy')[None,:]
                d_limit = 21201
            msb = True
            n_limit = 4294967296
            self._t_curr = 32
            compat_shift = self._t_curr-t if self._t_curr>=t else 0
            if compat_shift>0: warnings.warn("Truncating ints in generating matrix to have t = %d bits."%t,ParameterWarning)
            gen_mats = gen_mats>>compat_shift
        elif isinstance(generating_matrices,str):
            self.gen_mats_source = generating_matrices
            assert generating_matrices[-4:]==".txt"
            local_root = dirname(abspath(__file__))+'/generating_matrices/'
            repos = DataSource()
            if repos.exists(local_root+generating_matrices):
                datafile = repos.open(local_root+generating_matrices)
            elif repos.exists("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/dnet/"+generating_matrices):
                datafile = repos.open("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/dnet/"+generating_matrices)
            elif repos.exists("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/"+generating_matrices):
                datafile = repos.open("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/"+generating_matrices)
            elif repos.exists("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/dnet/"+generating_matrices[7:]):
                datafile = repos.open("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/dnet/"+generating_matrices[7:])
            elif repos.exists("https://raw.githubusercontent.com/QMCSoftware/"+generating_matrices):
                datafile = repos.open("https://raw.githubusercontent.com/QMCSoftware/"+generating_matrices)
            elif repos.exists(generating_matrices):
                datafile = repos.open(generating_matrices)
            else:
                raise ParameterError("LDData path %s not found"%generating_matrices)
            contents = [line.rstrip('\n').strip() for line in datafile.readlines()]
            contents = [line.split("#",1)[0] for line in contents if line[0]!="#"]
            datafile.close()
            msb = True
            assert int(contents[0])==2, "DigitalNetB2 requires base=2 " # base 2
            d_limit = int(contents[1])
            n_limit = int(contents[2])
            self._t_curr = int(contents[3])
            compat_shift = self._t_curr-t if self._t_curr>=t else 0
            if compat_shift>0: warnings.warn("Truncating ints in generating matrix to have t = %d bits."%t,ParameterWarning)
            gen_mats = np.array([[int(v)>>compat_shift for v in line.split(' ')] for line in contents[4:]],dtype=np.uint64)[None,:]
        elif isinstance(generating_matrices,np.ndarray):
            self.gen_mats_source = "custom"
            assert generating_matrices.ndim==2 or generating_matrices.ndim==3
            gen_mats = generating_matrices[None,:,:] if generating_matrices.ndim==2 else generating_matrices
            assert isinstance(msb,bool), "when generating_matrices is a np.ndarray you must set either msb=True (for most significant bit ordering) or msb=False (for least significant bit ordering which will require a bit reversal)"
            gen_mat_max = gen_mats.max() 
            assert gen_mat_max>0, "generating matrix must have positive ints"
            self._t_curr = int(np.ceil(np.log2(gen_mat_max+1)))
            d_limit = gen_mats.shape[1]
            n_limit = int(2**(gen_mats.shape[2]))
        else:
            raise ParameterError("invalid generating_matrices, must be a string or np.ndarray.")
        super(DigitalNetB2,self).__init__(dimension,replications,seed,d_limit,n_limit)
        assert gen_mats.ndim==3 and gen_mats.shape[1]>=self.d and (gen_mats.shape[0]==1 or gen_mats.shape[0]==self.replications) and gen_mats.shape[2]>0, "invalid gen_mats.shape = %s"%str(gen_mats.shape)
        self.m_max = int(gen_mats.shape[-1])
        if isinstance(generating_matrices,np.ndarray) and msb:
            qmctoolscl.dnb2_gmat_lsb_to_msb(np.uint64(gen_mats.shape[0]),np.uint64(self.d),np.uint64(self.m_max),np.tile(np.uint64(self._t_curr),int(gen_mats.shape[0])),gen_mats,gen_mats,backend="c")
        self.order = str(order).upper().strip().replace("_"," ")
        if self.order=="GRAY CODE": self.order = "GRAY"
        if self.order=="NATURAL": self.order = "RADICAL INVERSE"
        assert self.order in ['RADICAL INVERSE','GRAY']
        assert isinstance(t,int) and t>0
        assert self._t_curr<=t<=64, "t must no more than 64 and no less than %d (the number of bits used to represent the generating matrices)"%(self._t_curr)
        assert isinstance(alpha,int) and alpha>0
        self.alpha = alpha
        if self.alpha>1:
            assert (self.dvec==np.arange(self.d)).all(), "digital interlacing requires dimension is an int"
            if self.m_max!=self._t_curr:
                warnings.warn("Digital interlacing is often performed on matrices with the number of columns (m_max = %d) equal to the number of bits in each int (%d), but this is not the case. Ensure you are NOT setting alpha>1 when generating matrices are already interlaced."%(self.m_max,self._t_curr),ParameterWarning)
        self._verbose = _verbose
        self.randomize = str(randomize).upper().strip().replace("_"," ")
        if self.randomize=="TRUE": self.randomize = "LMS DS"
        if self.randomize=="OWEN": self.randomize = "NUS"
        if self.randomize=="NONE": self.randomize = "FALSE"
        if self.randomize=="NO": self.randomize = "FALSE"
        assert self.randomize in ["LMS DS","LMS","DS","NUS","FALSE"]
        self.dtalpha = self.alpha*self.d
        if self.randomize=="FALSE":
            if self.alpha==1:
                self.gen_mats = gen_mats[:,self.dvec,:]
                self.t = self._t_curr
            else: 
                t_alpha = min(self.alpha*self._t_curr,t)
                gen_mat_ho = np.empty((gen_mats.shape[0],self.d,self.m_max),dtype=np.uint64)
                qmctoolscl.dnb2_interlace(np.uint64(gen_mats.shape[0]),np.uint64(self.d),np.uint64(self.m_max),np.uint64(self.dtalpha),np.uint64(self._t_curr),np.uint64(t_alpha),np.uint64(self.alpha),gen_mats[:,:self.dtalpha,:].copy(),gen_mat_ho,backend="c")
                self.gen_mats = gen_mat_ho
                self._t_curr = t_alpha
                self.t = self._t_curr
        elif self.randomize=="DS":
            if self.alpha==1:
                self.gen_mats = gen_mats[:,self.dvec,:]
                self.t = t
            else: 
                t_alpha = min(self.alpha*self._t_curr,t)
                gen_mat_ho = np.empty((gen_mats.shape[0],self.d,self.m_max),dtype=np.uint64)
                qmctoolscl.dnb2_interlace(np.uint64(gen_mats.shape[0]),np.uint64(self.d),np.uint64(self.m_max),np.uint64(self.dtalpha),np.uint64(self._t_curr),np.uint64(t_alpha),np.uint64(self.alpha),gen_mats[:,:self.dtalpha,:].copy(),gen_mat_ho,backend="c")
                self.gen_mats = gen_mat_ho
                self._t_curr = t_alpha
                self.t = t
            self.rshift = qmctoolscl.random_tbit_uint64s(self.rng,self.t,(self.replications,self.d))
        elif self.randomize in ["LMS","LMS DS"]:
            if self.alpha==1:
                gen_mat_lms = np.empty((self.replications,self.d,self.m_max),dtype=np.uint64)
                S = qmctoolscl.dnb2_get_linear_scramble_matrix(self.rng,np.uint64(self.replications),np.uint64(self.d),np.uint64(self._t_curr),np.uint64(t),np.uint64(self._verbose))
                qmctoolscl.dnb2_linear_matrix_scramble(np.uint64(self.replications),np.uint64(self.d),np.uint64(self.m_max),np.uint64(gen_mats.shape[0]),np.uint64(t),S,gen_mats[:,self.dvec,:].copy(),gen_mat_lms,backend="c")
                self.gen_mats = gen_mat_lms
                self._t_curr = t
                self.t = self._t_curr
            else:
                gen_mat_lms = np.empty((self.replications,self.dtalpha,self.m_max),dtype=np.uint64)
                S = qmctoolscl.dnb2_get_linear_scramble_matrix(self.rng,np.uint64(self.replications),np.uint64(self.dtalpha),np.uint64(self._t_curr),np.uint64(t),np.uint64(self._verbose))
                qmctoolscl.dnb2_linear_matrix_scramble(np.uint64(self.replications),np.uint64(self.dtalpha),np.uint64(self.m_max),np.uint64(gen_mats.shape[0]),np.uint64(t),S,gen_mats[:,:self.dtalpha,:].copy(),gen_mat_lms,backend="c")
                gen_mat_lms_ho = np.empty((self.replications,self.d,self.m_max),dtype=np.uint64)
                qmctoolscl.dnb2_interlace(np.uint64(self.replications),np.uint64(self.d),np.uint64(self.m_max),np.uint64(self.dtalpha),np.uint64(t),np.uint64(t),np.uint64(self.alpha),gen_mat_lms,gen_mat_lms_ho,backend="c")
                self.gen_mats = gen_mat_lms_ho
                self._t_curr = t
                self.t = self._t_curr
            if self.randomize=="LMS DS":
                self.rshift = qmctoolscl.random_tbit_uint64s(self.rng,self.t,(self.replications,self.d))
        elif self.randomize=="NUS":
            if alpha==1:
                new_seeds = self._base_seed.spawn(self.replications*self.d)
                self.rngs = np.array([np.random.Generator(np.random.SFC64(new_seeds[j])) for j in range(self.replications*self.d)]).reshape(self.replications,self.d)
                self.root_nodes = np.array([qmctoolscl.NUSNode_dnb2() for i in range(self.replications*self.d)]).reshape(self.replications,self.d)
                self.gen_mats = gen_mats[:,self.dvec,:].copy()
                self.t = t
            else:
                self.dtalpha = self.alpha*self.d
                new_seeds = self._base_seed.spawn(self.replications*self.dtalpha)
                self.rngs = np.array([np.random.Generator(np.random.SFC64(new_seeds[j])) for j in range(self.replications*self.dtalpha)]).reshape(self.replications,self.dtalpha)
                self.root_nodes = np.array([qmctoolscl.NUSNode_dnb2() for i in range(self.replications*self.dtalpha)]).reshape(self.replications,self.dtalpha)
                self.gen_mats = gen_mats[:,:self.dtalpha,:].copy()
                self.t = t
        else:
            raise ParameterError("self.randomize parsing error")
        self.gen_mats = np.ascontiguousarray(self.gen_mats)
        gen_mat_max = self.gen_mats.max() 
        assert gen_mat_max>0, "generating matrix must have positive ints"
        assert self._t_curr==int(np.ceil(np.log2(gen_mat_max+1)))
        assert 0<self._t_curr<=self.t<=64, "invalid 0 <= self._t_curr (%d) <= self.t (%d) <= 64"%(self._t_curr,self.t)
        if self.randomize=="FALSE": assert self.gen_mats.shape[0]==self.replications, "randomize='FALSE' but replications = %d does not equal the number of sets of generating matrices %d"%(self.replications,self.gen_mats.shape[0])

    def _gen_samples(self, n_min, n_max, return_binary, warn):
        if n_min == 0 and self.randomize in ["FALSE","LMS"] and warn:
            warnings.warn("Without randomization, the first digtial net point is the origin",ParameterWarning)
        r_x = np.uint64(self.gen_mats.shape[0]) 
        n = np.uint64(n_max-n_min)
        d = np.uint64(self.dtalpha) if self.randomize=="NUS" and self.alpha>1 else np.uint64(self.d)
        n_start = np.uint64(n_min)
        mmax = np.uint64(self.m_max)
        xb = np.empty((r_x,n,d),dtype=np.uint64)
        if self.order=="GRAY":
            qmctoolscl.dnb2_gen_gray(r_x,n,d,n_start,mmax,self.gen_mats,xb,backend="c")
        elif self.order=="RADICAL INVERSE": 
            assert (n_min==0 or np.log2(n_min)%1==0) and (n_max==0 or np.log2(n_max)%1==0), "DigitalNetB2 in natural order requires n_min and n_max be 0 or powers of 2"
            qmctoolscl.dnb2_gen_natural(r_x,n,d,n_start,mmax,self.gen_mats,xb,backend="c")
        else:
            "invalid digital net order" 
        r = np.uint64(self.replications)
        if "NUS" in self.randomize:
            if self.alpha==1:
                xrb = np.empty((r,n,d),dtype=np.uint64)
                xb = np.ascontiguousarray(xb)
                qmctoolscl.dnb2_nested_uniform_scramble(r,n,d,r_x,np.uint64(self._t_curr),np.uint64(self.t),self.rngs,self.root_nodes,xb,xrb)
                xb = xrb
            else:
                d = np.uint64(self.d)
                dtalpha = np.uint64(self.dtalpha)
                t_alpha = self.t
                alpha = np.uint64(self.alpha)
                xrb_lo = np.empty((r,n,self.dtalpha),dtype=np.uint64)
                _ = qmctoolscl.dnb2_nested_uniform_scramble(r,n,dtalpha,r_x,np.uint64(self._t_curr),np.uint64(t_alpha),self.rngs,self.root_nodes,xb,xrb_lo)
                xrb_t = np.empty((r,d,n),dtype=np.uint64)
                xrb_lo_t = np.moveaxis(xrb_lo,[1,2],[2,1]).copy() 
                qmctoolscl.dnb2_interlace(r,d,n,dtalpha,np.uint64(t_alpha),np.uint64(self.t),alpha,xrb_lo_t,xrb_t)
                xb = np.moveaxis(xrb_t,[1,2],[2,1]).copy()
        if "DS" in self.randomize:
            xrb = np.empty((r,n,d),dtype=np.uint64)
            lshifts = np.tile(np.uint64((self.t-self._t_curr) if "LMS" not in self.randomize else 0),int(r))
            qmctoolscl.dnb2_digital_shift(r,n,d,r_x,lshifts,xb,self.rshift,xrb,backend="c")
            xb = xrb
        if return_binary:
            return xb
        else:
            x = np.empty((r,n,d),dtype=np.float64)
            tmaxes_new = np.tile(self.t,int(r)).astype(np.uint64)
            qmctoolscl.dnb2_integer_to_float(r,n,d,tmaxes_new,xb,x,backend="c")
            return x
    
    def _spawn(self, child_seed, dimension):
        return DigitalNetB2(
            dimension = dimension,
            replications = None if self.no_replications else self.replications,
            seed = child_seed,
            randomize = self.randomize,
            generating_matrices = self.input_generating_matrices,
            order = self.order,
            t = self.input_t,
            alpha = self.alpha,
            msb = self.input_msb,
            _verbose = False,
            # deprecated
            graycode = None,
            t_max = None,
            t_lms = None)



import numpy as np
from copy import copy


def _univ_repr(qmc_object, abc_class_name, attributes):
    """
    Clean way to represent qmc_object data.

    Args:
        qmc_object (object): an qmc_object instance
        abc_class_name (str): name of the abstract class
        attributes (list): list of attributes to include

    Returns:
        s (str): string representation of this qmcpy object

    Note:
        print(qmc_object) is equivalent to print(qmc_object.__repr__()). 
        See an abstract classes __repr__ method for example call to this method. 
    """
    with np.printoptions(precision=3,threshold=10):
        unique_attributes = []
        for attrib in attributes:
            if attrib not in unique_attributes:
                unique_attributes += [attrib]
        string = "%s (%s)" % (type(qmc_object).__name__, abc_class_name)
        for key in unique_attributes:
            val = getattr(qmc_object, key)
            # list of one value becomes just that value
            if isinstance(val, list) and len(val) == 1:
                val = copy(val[0])
            elif isinstance(val, list):
                try: val = np.array(val).copy()
                except: pass
            elif isinstance(val, np.ndarray):
                val = val.copy().squeeze()
                if val.shape == ():
                    val = val.item()
                elif val.size == 1:
                    val = val[0].item()
            # printing options
            s = '    %-15s '%key
            if isinstance(val, int) or isinstance(val,float): # scalar
                p = .1 if val<=0 else np.log2(float(val))
                if (p%1==0) and p!=0: # power of 2
                    s += '2^(%d)' % int(p)
                elif isinstance(val, int) or (val%1==0): # int
                    s += '%d' % int(val)
                else: # float 
                    if abs(val) < .001: # exponential format
                        s += '%.2e' % val
                    else: # float format
                        s += '%.3f' % val
            else:
                s += '%s' % str(val)
            string += '\n' + s.replace('\n', '\n     %-15s' % ' ')
    return string



from .abstract_stopping_criterion import AbstractStoppingCriterion
from ..discrete_distribution import DigitalNetB2
from ..integrand.ishigami import Ishigami
from ..true_measure.abstract_true_measure import AbstractTrueMeasure
from ..discrete_distribution.abstract_discrete_distribution import AbstractDiscreteDistribution
from ..util.data import Data

from ..util import ExactGPyTorchRegressionModel

from ..util import MaxSamplesWarning
import warnings
import time
import numpy as np
from scipy.stats import norm
import torch
import gpytorch


class Suggester(object): pass


class PFSampleErrorDensityAR(Suggester):
    def __init__(self, verbose=False):
        self.verbose = verbose
        super(PFSampleErrorDensityAR,self).__init__()
    def suggest(self, n, d, gp, rng, efficiency, pct=.5):
        if self.verbose: print('\tAR sampling with efficiency %.1e, expect %d draws: '%(efficiency,int(n/efficiency)),end='',flush=True)
        x = np.zeros((0,d),dtype=float)
        self.n_ar_tries_batch = 0
        rem = n
        draws_per_rem = int(np.ceil(np.log(pct)/np.log(1-efficiency)))
        while rem>0:
            newtries = rem*draws_per_rem
            self.n_ar_tries_batch += newtries
            if self.verbose: print('%d, '%self.n_ar_tries_batch,end='',flush=True)
            z = rng.random((newtries,d))
            u = rng.random(newtries)
            udens_z = _error_udens(gp,z)
            x = np.vstack([x,z[u<=udens_z]])
            rem = max(n-len(x),0)
        if self.verbose: print()
        return x[:n]


class SuggesterSimple(Suggester):
    def __init__(self, sampler):
        self.sampler = sampler
        if isinstance(self.sampler,AbstractTrueMeasure): assert (self.sampler.range==[0,1]).all()
        self.n_min = 0
        super(SuggesterSimple,self).__init__()
    def suggest(self, n, d, gp, rng, **kwargs):
        n_max = self.n_min+n
        assert d == self.sampler.d
        try: x = self.sampler(n_min=self.n_min,n_max=n_max)
        except: x = self.sampler(n)
        self.n_min = n_max
        return x


class PFGPCI(AbstractStoppingCriterion):
    """
    Probability of failure estimation using adaptive Gaussian process construction and resulting credible intervals. 
    
    Examples:
        >>> pfgpci = PFGPCI(
        ...     integrand = Ishigami(DigitalNetB2(3,seed=17)),
        ...     failure_threshold = 0,
        ...     failure_above_threshold = False, 
        ...     abs_tol = 2.5e-2,
        ...     alpha = 1e-1,
        ...     n_init = 64,
        ...     init_samples = None,
        ...     batch_sampler = PFSampleErrorDensityAR(verbose=False),
        ...     n_batch = 16,
        ...     n_limit = 128,
        ...     n_approx = 2**18,
        ...     gpytorch_prior_mean = gpytorch.means.ZeroMean(),
        ...     gpytorch_prior_cov = gpytorch.kernels.ScaleKernel(
        ...         gpytorch.kernels.MaternKernel(nu=2.5,lengthscale_constraint = gpytorch.constraints.Interval(.5,1)),
        ...         outputscale_constraint = gpytorch.constraints.Interval(1e-8,.5)),
        ...     gpytorch_likelihood = gpytorch.likelihoods.GaussianLikelihood(noise_constraint = gpytorch.constraints.Interval(1e-12,1e-8)),
        ...     gpytorch_marginal_log_likelihood_func = lambda likelihood,gpyt_model: gpytorch.mlls.ExactMarginalLogLikelihood(likelihood,gpyt_model),
        ...     torch_optimizer_func = lambda gpyt_model: torch.optim.Adam(gpyt_model.parameters(),lr=0.1),
        ...     gpytorch_train_iter = 100,
        ...     gpytorch_use_gpu = False,
        ...     verbose = False,
        ...     n_ref_approx = 2**22,
        ...     seed_ref_approx = 11)
        >>> solution,data = pfgpci.integrate(seed=7,refit=True)
        >>> data
        PFGPCIData (Data)
            solution        0.158
            error_bound     0.022
            bound_low       0.136
            bound_high      0.180
            n_total         112
            time_integrate  ...
        PFGPCI (AbstractStoppingCriterion)
            abs_tol         0.025
            n_init          2^(6)
            n_limit         2^(7)
            n_batch         2^(4)
        Ishigami (AbstractIntegrand)
        Uniform (AbstractTrueMeasure)
            lower_bound     -3.142
            upper_bound     3.142
        DigitalNetB2 (AbstractLDDiscreteDistribution)
            d               3
            replications    1
            randomize       LMS DS
            gen_mats_source joe_kuo.6.21201.txt
            order           RADICAL INVERSE
            t               63
            alpha           1
            n_limit         2^(32)
            entropy         17
        >>> df = data.get_results_dict()
        >>> with np.printoptions(formatter={"float": lambda x: "%-10.2e"%x, "int": lambda x: "%-10d"%x, "bool": lambda x: "%-10s"%x}):
        ...     for k,v in df.items():
        ...         print("%15s: %s"%(k,str(v)))
                   iter: [0          1          2          3         ]
                  n_sum: [64         80         96         112       ]
                n_batch: [64         16         16         16        ]
           error_bounds: [5.58e-02   3.92e-02   3.05e-02   2.16e-02  ]
                 ci_low: [8.66e-02   1.16e-01   1.19e-01   1.36e-01  ]
                ci_high: [1.98e-01   1.95e-01   1.80e-01   1.80e-01  ]
              solutions: [1.42e-01   1.55e-01   1.50e-01   1.58e-01  ]
          solutions_ref: [1.62e-01   1.62e-01   1.62e-01   1.62e-01  ]
              error_ref: [2.01e-02   7.02e-03   1.28e-02   4.52e-03  ]
                  in_ci: [True       True       True       True      ]
    
    **References:**
    
    1.  Sorokin, Aleksei G., and Vishwas Rao.  
        "Credible Intervals for Probability of Failure with Gaussian Processes."  
        arXiv preprint arXiv:2311.07733 (2023).
    """
    def __init__(self, 
        integrand,
        failure_threshold,
        failure_above_threshold,
        abs_tol = 5e-3,
        n_init = 64,
        n_limit = 1000,
        alpha = 1e-2,
        init_samples = None,
        batch_sampler = PFSampleErrorDensityAR(),
        n_batch = 4,
        n_approx = 2**20,
        gpytorch_prior_mean = gpytorch.means.ZeroMean(),
        gpytorch_prior_cov = gpytorch.kernels.ScaleKernel(gpytorch.kernels.MaternKernel(nu = 2.5)),
        gpytorch_likelihood = gpytorch.likelihoods.GaussianLikelihood(noise_constraint = gpytorch.constraints.Interval(1e-12,1e-8)),
        gpytorch_marginal_log_likelihood_func = lambda likelihood,gpyt_model: gpytorch.mlls.ExactMarginalLogLikelihood(likelihood,gpyt_model),
        torch_optimizer_func = lambda gpyt_model: torch.optim.Adam(gpyt_model.parameters(),lr=0.1),
        gpytorch_train_iter = 100,
        gpytorch_use_gpu = False,
        verbose = False,
        n_ref_approx = 2**22,
        seed_ref_approx = None):
        '''
        Args:
            integrand (AbstractIntegrand): The integrand.
            failure_threshold (float): Thresholds for failure. 
            failure_above_threshold (bool): Set to `True` if failure occurs when the simulation exceeds `failure_threshold` and False otherwise.
            abs_tol (float): The desired maximum distance from the estimate to either end of the credible interval.
            n_init (float): Initial number of samples from integrand.discrete_distrib from which to build the first surrogate GP
            n_limit (int): Budget of simulations.
            n_batch (int): The number of samples per batch to draw from batch_sampler. 
            alpha (float): The credible interval is constructed to hold with probability at least 1 - alpha
            init_samples (float): If the simulation has already been run, pass in (x,y) where x are past samples from the discrete distribution and y are corresponding simulation evaluations. 
            batch_sampler (Suggester or AbstractDiscreteDistribution): A suggestion scheme for future samples. 
            n_approx (int): Number of points from integrand.discrete_distrib used to approximate estimate and credible interval bounds
            gpytorch_prior_mean (gpytorch.means): prior mean function of the GP
            gpytorch_prior_cov (gpytorch.kernels): Prior covariance kernel of the GP
            gpytorch_likelihood (gpytorch.likelihoods): GP likelihood, require one of gpytorch.likelihoods.{GaussianLikelihood, GaussianLikelihoodWithMissingObs, FixedNoiseGaussianLikelihood}
            gpytorch_marginal_log_likelihood_func (callable): Function taking in the likelihood and gpytorch model and returning a marginal log likelihood from gpytorch.mlls
            torch_optimizer_func (callable): Function taking in the gpytorch model and returning an optimizer from torch.optim
            gpytorch_train_iter (int): Training iterations for the GP in gpytorch
            gpytorch_use_gpu (bool): If True, have gpytorch use a GPU for fitting and trining the GP
            verbose (int): If verbose > 0, print information through the call to integrate()
            n_ref_approx (int): If n_ref_approx > 0, use n_ref_approx points to get a reference QMC approximation of the true solution. 
                Caution: If n_ref_approx > 0, it should be a large int e.g. 2**22, in which case it is only helpful for cheap to evaluate simulations 
            seed_ref_approx (int): Seed for the reference approximation. Only applies when n_ref_approx>0
        '''
        self.parameters = ['abs_tol','n_init','n_limit','n_batch']
        self.integrand = integrand
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib
        self.d = self.integrand.d
        self.sampler = self.d
        self.failure_threshold = failure_threshold
        self.failure_above_threshold = failure_above_threshold
        self.abs_tol = abs_tol
        self.alpha = alpha
        assert 0<self.alpha<1
        self.n_init = n_init
        self.init_samples = init_samples is not None
        if self.init_samples: 
            self.x_init,self.y_init = init_samples
            assert self.x_init.ndim==2 and self.y_init.ndim==1
            assert self.x_init.shape[1]==self.d and len(self.y_init)==len(self.x_init)
            assert self.n_init==len(self.x_init)
            self.ytf_init = self._affine_tf(self.y_init)
        self.batch_sampler = batch_sampler
        self.n_batch = n_batch
        self.n_limit = n_limit
        assert self.n_limit >= self.n_init
        self.n_approx = n_approx
        assert (self.n_approx+self.n_init) <= 2**32
        self.gpytorch_prior_mean = gpytorch_prior_mean
        self.gpytorch_prior_cov = gpytorch_prior_cov
        self.gpytorch_likelihood = gpytorch_likelihood
        self.gpytorch_marginal_log_likelihood_func = gpytorch_marginal_log_likelihood_func
        self.torch_optimizer_func = torch_optimizer_func
        self.gpytorch_train_iter = gpytorch_train_iter
        self.gpytorch_use_gpu = gpytorch_use_gpu
        self.verbose = verbose
        self.approx_true_solution = n_ref_approx>0
        if self.approx_true_solution: 
            x = DigitalNetB2(self.d,order="GRAY",seed=seed_ref_approx)(n_ref_approx)
            y = self.integrand.f(x).squeeze()
            ytf = self._affine_tf(y)
            self.ref_approx = (ytf>=0).mean(0)
            if self.verbose: print('reference approximation with d=%d: %s'%(self.d,self.ref_approx))
        super(PFGPCI,self).__init__(allowed_distribs=[AbstractDiscreteDistribution],allow_vectorized_integrals=False)
    def _affine_tf(self, y):
        return y-self.failure_threshold if self.failure_above_threshold else self.failure_threshold-y
    def integrate(self, seed=None, refit=False):
        t0 = time.time()
        dnb2 = DigitalNetB2(self.d,randomize='DS',order="GRAY",seed=seed)
        data = PFGPCIData(self,self.integrand,self.true_measure,self.discrete_distrib,dnb2,self.n_approx,self.alpha,refit, 
                self.gpytorch_prior_mean,
                self.gpytorch_prior_cov,
                self.gpytorch_likelihood,
                self.gpytorch_marginal_log_likelihood_func,
                self.torch_optimizer_func,
                self.gpytorch_train_iter,
                self.gpytorch_use_gpu,
                self.verbose, self.approx_true_solution)
        batch_count = 0
        while True:
            if self.verbose: print('batch %d'%batch_count)
            if batch_count==0:
                if self.init_samples: 
                    xdraw,ydraw = self.x_init,self.y_init
                else:
                    xdraw = dnb2.spawn()[0](self.n_init)
                    ydraw = np.atleast_1d(self.integrand.f(xdraw).squeeze())
            else:
                n_new = min(self.n_batch,self.n_limit-np.sum(self.n_batch))
                xdraw = self.batch_sampler.suggest(n_new,self.d,data.gpyt_model,dnb2.rng,efficiency=2*data.emr[-1])
                ydraw = np.atleast_1d(self.integrand.f(xdraw).squeeze())
            ydrawtf = self._affine_tf(ydraw)
            data.update_data(batch_count, xdraw, ydrawtf)
            batch_count += 1
            if data.error_bounds[-1] <= self.abs_tol: break
            if sum(data.n_batch)==self.n_limit:
                warnings.warn('n_limit reached. ',MaxSamplesWarning)
                break        
        data.solution = data.solutions[-1]
        data.error_bound = data.error_bounds[-1]
        data.bound_low = data.ci_low[-1]
        data.bound_high = data.ci_high[-1]
        data.n_total = sum(data.n_batch)
        data.time_integrate = time.time()-t0
        data.n_sum = np.cumsum(data.n_batch)
        data.n_batch = np.array(data.n_batch)
        data.error_bounds = np.array(data.error_bounds)
        data.ci_low = np.array(data.ci_low)
        data.ci_high = np.array(data.ci_high)
        data.solutions = np.array(data.solutions)
        if self.approx_true_solution:
            data.solutions_ref = np.tile(self.ref_approx,len(data.n_batch))
            data.error_ref = abs(data.solutions-data.solutions_ref)
            data.in_ci = (data.ci_low<=data.solutions_ref)*(data.solutions_ref<=data.ci_high)
        return data.solution,data

def _get_phi(gp, x):
    yhat,yhatstd = gp.predict(x)
    with np.errstate(all='ignore'): z = yhat/yhatstd
    return norm.cdf(z)
def _error_udens_from_phi(phi):
    return 2*np.minimum(1-phi,phi)
def _error_udens(gp, x):
    return _error_udens_from_phi(_get_phi(gp,x))

class PFGPCIData(Data):
    """ Update and store data for the PFGPCI AbstractStoppingCriterion. """

    def __init__(self, stopping_crit, integrand, true_measure, discrete_distrib, dnb2, n_approx, alpha, refit, 
        gpytorch_prior_mean, gpytorch_prior_cov, gpytorch_likelihood, gpytorch_marginal_log_likelihood_func,
        torch_optimizer_func,gpytorch_train_iter,gpytorch_use_gpu,verbose,approx_true_solution):
        self.stopping_crit = stopping_crit
        self.integrand = integrand
        self.true_measure = true_measure
        self.discrete_distrib = discrete_distrib
        self.d = self.discrete_distrib.d
        self.dnb2 = dnb2
        self.alpha = alpha
        self.refit = refit
        self.qmc_pts = self.dnb2(n_approx)
        self.gpytorch_prior_mean = gpytorch_prior_mean
        self.gpytorch_prior_cov = gpytorch_prior_cov
        self.gpytorch_likelihood = gpytorch_likelihood
        self.gpytorch_marginal_log_likelihood_func = gpytorch_marginal_log_likelihood_func
        self.torch_optimizer_func = torch_optimizer_func
        self.gpytorch_train_iter = gpytorch_train_iter
        self.gpytorch_use_gpu = gpytorch_use_gpu
        self.verbose = verbose
        self.approx_true_solution = approx_true_solution
        self.n_batch = []
        self.emr = []
        self.error_bounds = []
        self.ci_low = []
        self.ci_high = []
        self.solutions = []
        self.x = np.empty((0,self.d),dtype=float)
        self.y = np.array([],dtype=float)
        self.gpyt_model = ExactGPyTorchRegressionModel(
            x_t = self.x, y_t = self.y,
            prior_mean = self.gpytorch_prior_mean,
            prior_cov = self.gpytorch_prior_cov,
            likelihood = self.gpytorch_likelihood,
            use_gpu = self.gpytorch_use_gpu)   
        self.saved_gps = [self.gpyt_model.state_dict()]
        super(PFGPCIData,self).__init__(parameters=['solution','error_bound','bound_low','bound_high','n_total','time_integrate'])

    def update_data(self, batch_count, xdraw, ydrawtf):
        self.n_batch.append(len(xdraw))
        self.x,self.y = np.vstack([self.x,xdraw]),np.hstack([self.y,ydrawtf])
        if batch_count==0 or self.refit:
            self.gpyt_model = ExactGPyTorchRegressionModel(
                x_t = self.x, y_t = self.y,
                prior_mean = self.gpytorch_prior_mean,
                prior_cov = self.gpytorch_prior_cov,
                likelihood = self.gpytorch_likelihood,
                use_gpu = self.gpytorch_use_gpu)
            self.gpyt_model.fit(
                optimizer = self.torch_optimizer_func(self.gpyt_model),
                mll = self.gpytorch_marginal_log_likelihood_func(self.gpyt_model.likelihood,self.gpyt_model),
                training_iter = self.gpytorch_train_iter,
                verbose = self.verbose)
        else:
            self.gpyt_model = self.gpyt_model.add_data(xdraw,ydrawtf)
            torch.cuda.empty_cache()          
        self.saved_gps.append(self.gpyt_model.state_dict())
        phi = _get_phi(self.gpyt_model,self.qmc_pts)
        self.solutions.append(np.mean(phi>=.5))
        self.emr.append(np.mean(np.minimum(phi,1-phi)))
        gamma = self.emr[-1]/self.alpha
        self.ci_low.append(np.maximum(self.solutions[-1]-gamma,0))
        self.ci_high.append(np.minimum(self.solutions[-1]+gamma,1))
        self.error_bounds.append(np.maximum(self.solutions[-1]-self.ci_low[-1],self.ci_high[-1]-self.solutions[-1]))

    def get_results_dict(self):
        df = {'iter':np.arange(len(self.n_sum)), 'n_sum':self.n_sum, 'n_batch':self.n_batch, 'error_bounds':self.error_bounds, 'ci_low':self.ci_low, 'ci_high':np.array(self.ci_high), 'solutions':np.array(self.solutions)}
        if self.approx_true_solution: df['solutions_ref'],df['error_ref'],df['in_ci'] = self.solutions_ref,self.error_ref,self.in_ci
        return df 
    
    def plot(self, trace_only=False, **kwargs):
        from matplotlib import pyplot
        if self.d==1 and not trace_only and self.saved_gps!=[]: fig,gs = self.plot_1d(**kwargs)
        elif self.d==2 and not trace_only and self.saved_gps!=[]: fig,gs = self.plot_2d(**kwargs)
        else:
            from matplotlib import pyplot,gridspec
            fig = pyplot.figure(constrained_layout=False,figsize=(8,4))
            gs = gridspec.GridSpec(1,2,figure=fig)
        axtrace = fig.add_subplot(gs[-1,:])
        if self.approx_true_solution: axtrace.plot(self.n_sum,self.solutions_ref,color='c',label=r'$P(g)$')
        axtrace.plot(self.n_sum,self.solutions,'-o',color='k',label=r'$\hat{P}_n^\mathrm{QMC}$')
        axtrace.fill_between(self.n_sum,self.ci_low,self.ci_high,color='k',alpha=.15)
        axtrace.plot(self.n_sum,self.error_bounds,'-o',color='m',label=r'$\hat{\gamma}_n^{\text{QMC}}$')
        if self.approx_true_solution: axtrace.plot(self.n_sum,self.error_ref,'-o',color='b',label=r'$\vert \hat{P}_n^{\text{QMC}} - P(g) \vert$')
        axtrace.set_xlim([self.n_sum[0]-self.n_batch[1]/2,self.n_sum[-1]+self.n_batch[1]/2])
        # axtrace.set_xticks(self.n_sum)
        # axtrace.set_xticklabels(self.n_sum)
        axtrace.set_yscale('log',base=10)
        #axtrace.set_ylim([0,1]); axtrace.set_yticks([0,1])
        for spine in ['top','right']: axtrace.spines[spine].set_visible(False)
        axtrace.legend(loc='lower center',bbox_to_anchor=(0.05, -.5, .9, .102),mode='expand',ncol=4,frameon=False)
        return fig
    def plot_1d(self, meshticks=1025, ci_percentage=.95, **kwargs):
        from matplotlib import pyplot,gridspec
        beta = norm.ppf(np.mean([ci_percentage,1]))
        n_batches = len(self.n_batch)
        fig = pyplot.figure(constrained_layout=False,figsize=(5*n_batches,5*3))
        gs = gridspec.GridSpec(3,n_batches,figure=fig)
        xticks = np.linspace(0,1,meshticks)[1:-1]
        if self.approx_true_solution: 
            yticks = self.integrand.f(xticks[:,None]).squeeze()
            ytickstf = self.stopping_crit._affine_tf(yticks)
        for j in range(n_batches):
            i0,i1 = 0 if j==0 else self.n_sum[j-1],self.n_sum[j]
            gpyt_model = ExactGPyTorchRegressionModel(
                x_t = self.x[:i0], y_t = self.y[:i0],
                prior_mean = self.gpytorch_prior_mean,
                prior_cov = self.gpytorch_prior_cov,
                likelihood = self.gpytorch_likelihood,
                use_gpu = self.gpytorch_use_gpu)
            gpyt_model.load_state_dict(self.saved_gps[j])
            ax0j = fig.add_subplot(gs[0,j])
            mr_udens = _error_udens(gpyt_model,xticks[:,None])
            ax0j.fill_between(xticks,mr_udens,color='k',alpha=.5)
            ax0j.scatter(self.x[i0:i1].squeeze(),np.zeros(i1-i0)+.015,color='r')
            ax0j.set_ylim([0,1])
            ax0j.set_yticks([0,1])
            gpyt_model = ExactGPyTorchRegressionModel(
                x_t = self.x[:i1], y_t = self.y[:i1],
                prior_mean = self.gpytorch_prior_mean,
                prior_cov = self.gpytorch_prior_cov,
                likelihood = self.gpytorch_likelihood,
                use_gpu = self.gpytorch_use_gpu)
            gpyt_model.load_state_dict(self.saved_gps[j+1])
            ax1j = fig.add_subplot(gs[1,j],sharey=None if j==0 else ax1j)
            if self.approx_true_solution: ax1j.plot(xticks,ytickstf,color='c',label=r'f(x)')
            gp_mean,gp_std = gpyt_model.predict(xticks[:,None])
            ax1j.plot(xticks,gp_mean,color='k',label='gp')
            ax1j.fill_between(xticks,gp_mean-beta*gp_std,gp_mean+beta*gp_std,color='k',alpha=.25)
            ax1j.axhline(y=0,color='lightgreen')
            ax1j.scatter(self.x[:i0].squeeze(),self.y[:i0],color='b',label='old pts')
            ax1j.scatter(self.x[i0:i1].squeeze(),self.y[i0:i1],color='r',label='new pts')
            # ax1j.set_xlabel(r'$u$')
            if j==0:
                #ax1j.set_ylabel(r'$y$')
                ax0j.set_ylabel(r'$2\mathrm{ERR}_n(u)$')
                ax1j.set_ylabel('GP Regression')
        for ax in fig.axes:
            ax.set_xlim([0,1])
            ax.set_xticks([0,1])
            ax.xaxis.set_visible(False)
        return fig,gs
    def plot_2d(self, meshticks=257, clevels=32, **kwargs):
        from matplotlib import pyplot,gridspec,cm
        n_batches = len(self.n_batch)
        fig = pyplot.figure(constrained_layout=False,figsize=(5*n_batches,5*5))
        gs = gridspec.GridSpec(5 if self.approx_true_solution else 4,n_batches,figure=fig)
        x0mesh,x1mesh = np.meshgrid(np.linspace(0,1,meshticks)[1:-1],np.linspace(0,1,meshticks)[1:-1])
        xquery = np.vstack([x0mesh.flatten(),x1mesh.flatten()]).T
        if self.approx_true_solution:
            yquery = self.integrand.f(xquery).squeeze()
            ymeshtf = self.stopping_crit._affine_tf(yquery).reshape(x0mesh.shape)
        for j in range(n_batches):
            row_idx = 0
            i0,i1 = 0 if j==0 else self.n_sum[j-1],self.n_sum[j]
            if self.approx_true_solution:
                ax0j = fig.add_subplot(gs[row_idx,j])
                ax0j.contourf(x0mesh,x1mesh,ymeshtf,cmap=cm.Greys,vmin=ymeshtf.min(),vmax=ymeshtf.max(),levels=clevels)
                ax0j.contour(x0mesh,x1mesh,ymeshtf,colors='lightgreen',levels=[0])
                # ax0j.set_title(r'$g(\boldsymbol{u})$')
                if j==0:
                    ax0j.set_ylabel(r'$g(\boldsymbol{u})$')
                row_idx += 1
            gpyt_model = ExactGPyTorchRegressionModel(
                x_t = self.x[:i0], y_t = self.y[:i0],
                prior_mean = self.gpytorch_prior_mean,
                prior_cov = self.gpytorch_prior_cov,
                likelihood = self.gpytorch_likelihood,
                use_gpu = self.gpytorch_use_gpu)
            gpyt_model.load_state_dict(self.saved_gps[j])
            ax1j = fig.add_subplot(gs[row_idx,j])
            udens_mr = _error_udens(gpyt_model,xquery).reshape(x0mesh.shape)
            ax1j.contourf(x0mesh,x1mesh,udens_mr,cmap=cm.Greys,levels=clevels,vmin=0,vmax=1)
            ax1j.scatter(self.x[i0:i1,0],self.x[i0:i1,1],color='r')
            # ax1j.set_title(r'$2\mathrm{ERR}_n(\boldsymbol{u})$')
            if j==0:
                ax1j.set_ylabel(r'$2\mathrm{ERR}_n(\boldsymbol{u})$')
            row_idx += 1
            gpyt_model = ExactGPyTorchRegressionModel(
                x_t = self.x[:i0], y_t = self.y[:i0],
                prior_mean = self.gpytorch_prior_mean,
                prior_cov = self.gpytorch_prior_cov,
                likelihood = self.gpytorch_likelihood,
                use_gpu = self.gpytorch_use_gpu)
            gpyt_model.load_state_dict(self.saved_gps[j+1])
            gp_mean,gp_std = gpyt_model.predict(xquery)
            gp_mean_mesh,gp_std_mesh = gp_mean.reshape(x0mesh.shape),gp_std.reshape(x0mesh.shape)
            ax2j = fig.add_subplot(gs[row_idx,j])
            ax2j.contourf(x0mesh,x1mesh,gp_mean_mesh,cmap=cm.Greys,vmin=gp_mean.min(),vmax=gp_mean.max(),levels=clevels)
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                ax2j.contour(x0mesh,x1mesh,gp_mean_mesh,colors='lightgreen',levels=[0])
            # ax2j.set_title(r'$M_n(\boldsymbol{u})$')
            if j==0:
                ax2j.set_ylabel(r'$M_n(\boldsymbol{u})$')
            row_idx += 1
            ax3j = fig.add_subplot(gs[row_idx,j])            
            ax3j.contourf(x0mesh,x1mesh,gp_std_mesh,cmap=cm.Greys,levels=clevels)
            # ax3j.set_title(r'$\sigma_n(\boldsymbol{u})$')
            if j==0:
                ax3j.set_ylabel(r'$\sigma_n(\boldsymbol{u})$')
            for ax in [ax2j,ax3j]:
                ax.scatter(self.x[:i0,0],self.x[:i0,1],color='b')
                ax.scatter(self.x[i0:i1,0],self.x[i0:i1,1],color='r')
        for ax in fig.axes:
            ax.set_xlim([0,1])
            ax.set_ylim([0,1])
            # ax.set_xlabel(r'$u_1$')
            ax.set_aspect(1)
            ax.xaxis.set_visible(False)
            ax.yaxis.set_tick_params(top=False, bottom=False, left=False, right=False,
                labelleft=True, labelbottom=False)
            # ax.set_xticks([0,1])
            # ax.set_yticks([0,1])
            # ax.set_ylabel(r'$u_2$')
        return fig,gs



from .abstract_true_measure import AbstractTrueMeasure
from ..util import DimensionError, ParameterError
from ..discrete_distribution import DigitalNetB2
import numpy as np
from typing import Union


class Kumaraswamy(AbstractTrueMeasure):
    r"""
    Kumaraswamy distribution as described in [https://en.wikipedia.org/wiki/Kumaraswamy_distribution](https://en.wikipedia.org/wiki/Kumaraswamy_distribution).
    
    Examples:
        >>> true_measure = Kumaraswamy(DigitalNetB2(2,seed=7),a=[1,2],b=[3,4])
        >>> true_measure(4)
        array([[0.34705366, 0.6782161 ],
               [0.0577568 , 0.36189538],
               [0.76344358, 0.0932949 ],
               [0.17065545, 0.43009386]])
        >>> true_measure
        Kumaraswamy (AbstractTrueMeasure)
            a               [1 2]
            b               [3 4]
        
        With independent replications 

        >>> x = Kumaraswamy(DigitalNetB2(3,seed=7,replications=2),a=[1,2,3],b=[3,4,5])(4)
        >>> x.shape 
        (2, 4, 3)
        >>> x
        array([[[0.09004177, 0.22144305, 0.62190133],
                [0.31710078, 0.48718217, 0.47325643],
                [0.19657641, 0.57423463, 0.25697057],
                [0.56103074, 0.28939035, 0.63654112]],
        <BLANKLINE>
               [[0.18006788, 0.62226635, 0.5083556 ],
                [0.22602452, 0.10519477, 0.42823814],
                [0.08428482, 0.28804621, 0.2414302 ],
                [0.37253319, 0.45379743, 0.63366422]]])
    """

    def __init__(self, sampler, a=2, b=2):
        r"""
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
            a (Union[float,np.ndarray]): First parameter $\alpha > 0$.
            b (Union[float,np.ndarray]): Second parameter $\beta > 0$.
        """
        self.parameters = ['a', 'b']
        self.domain = np.array([[0,1]])
        self.range = np.array([[0,1]])
        self._parse_sampler(sampler)
        self.a = a
        self.b = b
        self.alpha = np.array(a)
        if self.alpha.size==1:
            self.alpha = self.alpha.item()*np.ones(self.d)
            a = np.tile(self.a,self.d)
        self.beta = np.array(b)
        if self.beta.size==1:
            self.beta = self.beta.item()*np.ones(self.d)
        if not (self.alpha.shape==(self.d,) and self.beta.shape==(self.d,)):
            raise DimensionError('a and b must be scalar or have length equal to dimension.')
        if not ((self.alpha>0).all() and (self.beta>0).all()):
            raise ParameterError("Kumaraswamy requires a,b>0.")
        super(Kumaraswamy,self).__init__()
        assert self.alpha.shape==(self.d,) and self.beta.shape==(self.d,)

    def _transform(self, x):
        return (1-(1-x)**(1/self.beta))**(1/self.alpha)
    
    def _weight(self, x):
        return np.prod(self.alpha*self.beta*x**(self.alpha-1)*(1-x**self.alpha)**(self.beta-1),-1)
    
    def _spawn(self, sampler, dimension):
        if dimension==self.d: # don't do anything if the dimension doesn't change
            spawn = Kumaraswamy(sampler,a=self.alpha,b=self.beta)
        else:
            a = self.alpha[0]
            b = self.beta[0]
            if not (all(self.alpha==a) and all(self.beta==b)):
                raise DimensionError('''
                    In order to spawn a Kumaraswamy measure
                    a must all be the same and 
                    b must all be the same''')
            spawn = Kumaraswamy(sampler,a=a,b=b)
        return spawn



import qmctoolscl 
import numpy as np 

def _parse_ft_input(x):
    if x.ndim==1:
        d,n = 1,len(x)
        x = x[None,:]
        shape = (n,)
    elif x.ndim==2:
        d,n = x.shape
        shape = (d,n)
    else:
        shape = x.shape
        n = shape[-1]
        x = x.reshape(-1,n)
        d = x.shape[0]
    assert (n&(n-1))==0 # require n is 0 or a power of 2
    return x,shape,d,n,n//2

def fftbr_qmctoolscl(x):
    r"""
    QMCToolsCL implementation of the 1 dimensional Bit-Reversed-Order (BRO) Fast Fourier Transform (FFT) along the last dimension. 
    Requires the last dimension of x is already in BRO, so we can skip the first step of the decimation-in-time FFT. 
    Requires the size of the last dimension is a power of 2. 

    Examples:
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> x = rng.random(8)+1j*rng.random(8)
        >>> with np.printoptions(precision=2):
        ...     fftbr_qmctoolscl(x)
        array([ 1.07+1.14j, -0.32+0.26j, -0.27+0.22j, -0.27-0.34j,  0.1 +0.16j,
               -0.15+0.17j,  0.5 +0.24j,  0.06-0.02j])
        >>> fftbr_qmctoolscl(rng.random((2,3,4,5,8))+1j*rng.random((2,3,4,5,8))).shape
        (2, 3, 4, 5, 8)

    Args:
        x (np.ndarray): Array of samples at which to run BRO-FFT.
    
    Returns:
        y (np.ndarray): BRO-FFT values.
    """
    x,shape,d,n,n_half = _parse_ft_input(x)
    if n<=1: return x.reshape(shape).copy()+0*1j
    twiddler = np.empty(n,dtype=np.float64)
    twiddlei = np.empty(n,dtype=np.float64)
    xr = x.real.copy()
    xi = x.imag.copy()
    qmctoolscl.fft_bro_1d_radix2(1,d,int(n_half),twiddler,twiddlei,xr,xi)
    xc = xr+1j*xi
    return xc.reshape(shape)

def ifftbr_qmctoolscl(x):
    r"""
    QMCToolsCL implementation of the 1 dimensional Bit-Reversed-Order (BRO) Inverse Fast Fourier Transform (IFFT) along the last dimension.  
    Outputs an array in bit-reversed order, so we can skip the last step of the decimation-in-time IFFT.  
    Requires the size of the last dimension is a power of 2. 

    Examples:
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> x = rng.random(8)+1j*rng.random(8)
        >>> with np.printoptions(precision=2):
        ...     ifftbr_qmctoolscl(x)
        array([ 1.07+1.14j, -0.29-0.22j,  0.3 +0.06j, -0.09+0.02j,  0.03+0.54j,
               -0.04-0.33j, -0.19+0.26j, -0.08+0.36j])
        >>> ifftbr_qmctoolscl(rng.random((2,3,4,5,8))+1j*rng.random((2,3,4,5,8))).shape
        (2, 3, 4, 5, 8)
    
    Args:
        x (np.ndarray): Array of samples at which to run BRO-IFFT.
    
    Returns:
        y (np.ndarray): BRO-IFFT values.
    """
    x,shape,d,n,n_half = _parse_ft_input(x)
    if n<=1: return x.reshape(shape).copy()+0*1j
    twiddler = np.empty(n,dtype=np.float64)
    twiddlei = np.empty(n,dtype=np.float64)
    xr = x.real.copy()
    xi = x.imag.copy()
    qmctoolscl.ifft_bro_1d_radix2(1,d,int(n_half),twiddler,twiddlei,xr,xi)
    xc = xr+1j*xi
    return xc.reshape(shape)

def fwht_qmctoolscl(x):
    r"""
    QMCToolsCL implementation of the 1 dimensional Fast Walsh Hadamard Transform (FWHT) along the last dimension.  
    Requires the size of the last dimension is a power of 2. 

    Examples:
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> with np.printoptions(precision=2):
        ...     fwht_qmctoolscl(rng.random(8))
        array([ 1.07, -0.29,  0.12,  0.08,  0.1 , -0.45,  0.1 , -0.03])
        >>> fwht_qmctoolscl(rng.random((2,3,4,5,8))).shape
        (2, 3, 4, 5, 8)

    Args:
        x (np.ndarray): Array of samples at which to run FWHT.
    
    Returns:
        y (np.ndarray): FWHT values.
    """
    x,shape,d,n,n_half = _parse_ft_input(x)
    if n<=1: return x.reshape(shape).copy()
    xcp = x.copy()
    qmctoolscl.fwht_1d_radix2(1,d,int(n_half),xcp)
    return xcp.reshape(shape)



from .abstract_true_measure import AbstractTrueMeasure
from .brownian_motion import BrownianMotion
from .geometric_brownian_motion import GeometricBrownianMotion
from .gaussian import Gaussian
from .lebesgue import Lebesgue
from .uniform import Uniform
from .kumaraswamy import Kumaraswamy
from .bernoulli_cont import BernoulliCont
from .johnsons_su import JohnsonsSU
from .scipy_wrapper import SciPyWrapper
from .matern_gp import MaternGP

TrueMeasure = AbstractTrueMeasure
_TrueMeasure = AbstractTrueMeasure
Normal = Gaussian
Matern = MaternGP


from .abstract_stopping_criterion import AbstractStoppingCriterion
import numpy as np
from scipy.stats import norm

class AbstractCubMLMC(AbstractStoppingCriterion):
    
    def _get_next_samples(self, data):
        ns = np.ceil( np.sqrt(data.var_level/data.cost_per_sample) * 
                np.sqrt(data.var_level*data.cost_per_sample).sum() / 
                ((1-self.theta)*self.rmse_tol**2) )
        return ns.astype(int)
    
    def set_tolerance(self, abs_tol=None, rel_tol=None, rmse_tol=None):
        assert rel_tol is None, "rel_tol not supported by this stopping criterion."
        if rmse_tol != None:
            self.rmse_tol = float(rmse_tol)
        elif abs_tol != None:
            self.rmse_tol = (float(abs_tol) / norm.ppf(1-self.alpha/2.))
    
    def _update_data(self, data):
        for l in range(data.levels+1):
            if l==len(data.level_integrands):
                # haven't spawned this level's integrand yet
                data.level_integrands += self.integrand.spawn(levels=int(l))
            integrand_l = data.level_integrands[l]
            if data.diff_n_level[l] > 0:
                # evaluate integral at sampling points samples
                n = data.diff_n_level[l]
                samples = integrand_l.discrete_distrib(n=n)
                pc,pf = integrand_l.f(samples)
                dp = pf-pc
                data.n_level[l] = data.n_level[l] + data.diff_n_level[l]
                data.sum_level[0,l] = data.sum_level[0,l] + dp.sum()
                data.sum_level[1,l] = data.sum_level[1,l] + (dp**2).sum()
                data.cost_level[l] = data.cost_level[l] + integrand_l.cost*n
        # compute absolute average, variance and cost
        data.mean_level = np.absolute(data.sum_level[0,:data.levels+1]/data.n_level[:data.levels+1])
        data.var_level = np.maximum(0,data.sum_level[1,:data.levels+1]/data.n_level[:data.levels+1] - data.mean_level**2)
        data.cost_per_sample = data.cost_level[:data.levels+1]/data.n_level[:data.levels+1]
        # fix to cope with possible zero values for data.mean_level and data.var_level
        # (can happen in some applications when there are few samples)
        for l in range(2,data.levels+1):
            data.mean_level[l] = np.maximum(data.mean_level[l], .5*data.mean_level[l-1]/2**data.alpha)
            data.var_level[l] = np.maximum(data.var_level[l], .5*data.var_level[l-1]/2**data.beta)
        # use linear regression to estimate alpha, beta, gamma if not given
        a = np.ones((data.levels,2))
        a[:,0] = np.arange(1,data.levels+1)
        if self.alpha0 <= 0:
            x = np.linalg.lstsq(a,np.log2(data.mean_level[1:]),rcond=None)[0]
            data.alpha = np.maximum(.5,-x[0])
        if self.beta0 <= 0:
            x = np.linalg.lstsq(a,np.log2(data.var_level[1:]),rcond=None)[0]
            data.beta = np.maximum(.5,-x[0])
        if self.gamma0 <= 0:
            x = np.linalg.lstsq(a,np.log2(data.cost_per_sample[1:]),rcond=None)[0]
            data.gamma = np.maximum(.5,x[0])
        data.n_total = data.n_level.sum()

    def _add_level(self, data):
        data.levels += 1
        if not len(data.n_level) > data.levels:
            data.mean_level = np.hstack((data.mean_level, data.mean_level[-1] / 2**data.alpha))
            data.var_level = np.hstack((data.var_level, data.var_level[-1] / 2**data.beta))
            data.cost_per_sample = np.hstack((data.cost_per_sample, data.cost_per_sample[-1] * 2**data.gamma))
            data.n_level = np.hstack((data.n_level, 0))
            data.sum_level = np.hstack((data.sum_level,np.zeros((2,1))))
            data.cost_level = np.hstack((data.cost_level, 0.))
        else:
            data.mean_level = np.absolute(data.sum_level[0,:data.levels+1]/data.n_level[:data.levels+1])
            data.var_level = np.maximum(0,data.sum_level[1,:data.levels+1]/data.n_level[:data.levels+1] - data.mean_level**2)
            data.cost_per_sample = data.cost_level[:data.levels+1]/data.n_level[:data.levels+1]


from .abstract_cub_qmc_ld_g import AbstractCubQMCLDG
from ..discrete_distribution import Lattice
from ..true_measure import Gaussian, Uniform
from ..integrand import Keister, BoxIntegral, CustomFun
from ..integrand.genz import Genz
from ..integrand.sensitivity_indices import SensitivityIndices
from ..fast_transform import fftbr,omega_fftbr
from ..util import ParameterError
import numpy as np


class CubQMCLatticeG(AbstractCubQMCLDG):
    r"""
    Quasi-Monte Carlo stopping criterion using rank-1 lattice cubature 
    with guarantees for cones of functions with a predictable decay in the Fourier coefficients.
    
    Examples:
        >>> k = Keister(Lattice(seed=7))
        >>> sc = CubQMCLatticeG(k,abs_tol=1e-3,rel_tol=0,check_cone=True)
        >>> solution,data = sc.integrate()
        >>> solution
        array(1.38037385)
        >>> data
        Data (Data)
            solution        1.380
            comb_bound_low  1.380
            comb_bound_high 1.381
            comb_bound_diff 0.001
            comb_flags      1
            n_total         2^(11)
            n               2^(11)
            time_integrate  ...
        CubQMCLatticeG (AbstractStoppingCriterion)
            abs_tol         0.001
            rel_tol         0
            n_init          2^(10)
            n_limit         2^(30)
        Keister (AbstractIntegrand)
        Gaussian (AbstractTrueMeasure)
            mean            0
            covariance      2^(-1)
            decomp_type     PCA
        Lattice (AbstractLDDiscreteDistribution)
            d               1
            replications    1
            randomize       SHIFT
            gen_vec_source  kuo.lattice-33002-1024-1048576.9125.txt
            order           RADICAL INVERSE
            n_limit         2^(20)
            entropy         7
        
        Vector outputs
        
        >>> f = BoxIntegral(Lattice(3,seed=11),s=[-1,1])
        >>> abs_tol = 1e-3
        >>> sc = CubQMCLatticeG(f,abs_tol=abs_tol,rel_tol=0,check_cone=True)
        >>> solution,data = sc.integrate()
        >>> solution
        array([1.18947477, 0.96060862])
        >>> data
        Data (Data)
            solution        [1.189 0.961]
            comb_bound_low  [1.189 0.96 ]
            comb_bound_high [1.19  0.961]
            comb_bound_diff [0.001 0.001]
            comb_flags      [ True  True]
            n_total         2^(13)
            n               [8192 1024]
            time_integrate  ...
        CubQMCLatticeG (AbstractStoppingCriterion)
            abs_tol         0.001
            rel_tol         0
            n_init          2^(10)
            n_limit         2^(30)
        BoxIntegral (AbstractIntegrand)
            s               [-1  1]
        Uniform (AbstractTrueMeasure)
            lower_bound     0
            upper_bound     1
        Lattice (AbstractLDDiscreteDistribution)
            d               3
            replications    1
            randomize       SHIFT
            gen_vec_source  kuo.lattice-33002-1024-1048576.9125.txt
            order           RADICAL INVERSE
            n_limit         2^(20)
            entropy         11
        >>> sol3neg1 = -np.pi/4-1/2*np.log(2)+np.log(5+3*np.sqrt(3))
        >>> sol31 = np.sqrt(3)/4+1/2*np.log(2+np.sqrt(3))-np.pi/24
        >>> true_value = np.array([sol3neg1,sol31])
        >>> assert (abs(true_value-solution)<abs_tol).all()

        Sensitivity indices 

        >>> function = Genz(Lattice(3,seed=7))
        >>> integrand = SensitivityIndices(function)
        >>> sc = CubQMCLatticeG(integrand,abs_tol=5e-4,rel_tol=0,check_cone=True)
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        [[0.021 0.196 0.667]
                             [0.036 0.303 0.782]]
            comb_bound_low  [[0.02  0.196 0.667]
                             [0.035 0.303 0.781]]
            comb_bound_high [[0.021 0.196 0.667]
                             [0.036 0.303 0.782]]
            comb_bound_diff [[0.001 0.001 0.   ]
                             [0.001 0.001 0.001]]
            comb_flags      [[ True  True  True]
                             [ True  True  True]]
            n_total         2^(16)
            n               [[[16384 32768 65536]
                              [16384 32768 65536]
                              [16384 32768 65536]]
        <BLANKLINE>                
                             [[ 2048 16384 32768]
                              [ 2048 16384 32768]
                              [ 2048 16384 32768]]]
            time_integrate  ...
        CubQMCLatticeG (AbstractStoppingCriterion)
            abs_tol         5.00e-04
            rel_tol         0
            n_init          2^(10)
            n_limit         2^(30)
        SensitivityIndices (AbstractIntegrand)
            indices         [[ True False False]
                             [False  True False]
                             [False False  True]]
        Uniform (AbstractTrueMeasure)
            lower_bound     0
            upper_bound     1
        Lattice (AbstractLDDiscreteDistribution)
            d               3
            replications    1
            randomize       SHIFT
            gen_vec_source  kuo.lattice-33002-1024-1048576.9125.txt
            order           RADICAL INVERSE
            n_limit         2^(20)
            entropy         7
    
    **References:**

    1.  Lluis Antoni Jimenez Rugama and Fred J. Hickernell.  
        "Adaptive multidimensional integration based on rank-1 lattices,"   
        Monte Carlo and Quasi-Monte Carlo Methods: MCQMC, Leuven, Belgium,  
        April 2014 (R. Cools and D. Nuyens, eds.), Springer Proceedings in Mathematics.  
        and Statistics, vol. 163, Springer-Verlag, Berlin, 2016, arXiv:1411.1966, pp. 407-422.
        
    2.  Sou-Cheng T. Choi, Yuhan Ding, Fred J. Hickernell, Lan Jiang, Lluis Antoni Jimenez Rugama,  
        Da Li, Jagadeeswaran Rathinavel, Xin Tong, Kan Zhang, Yizhi Zhang, and Xuan Zhou,  
        GAIL: Guaranteed Automatic Integration Library (Version 2.3) [MATLAB Software], 2019.  
        [http://gailgithub.github.io/GAIL_Dev/](http://gailgithub.github.io/GAIL_Dev/).  
        [https://github.com/GailGithub/GAIL_Dev/blob/master/Algorithms/IntegrationExpectation/cubLattice_g.m](https://github.com/GailGithub/GAIL_Dev/blob/master/Algorithms/IntegrationExpectation/cubLattice_g.m).
    """

    def __init__(self, 
                 integrand, 
                 abs_tol = 1e-2, 
                 rel_tol = 0., 
                 n_init = 2**10,
                 n_limit = 2**30,
                 error_fun = "EITHER",
                 fudge = lambda m: 5.*2.**(-m), 
                 check_cone = False,
                 ptransform = 'BAKER',
                 ):
        r"""
        Args:
            integrand (AbstractIntegrand): The integrand.
            abs_tol (np.ndarray): Absolute error tolerance.
            rel_tol (np.ndarray): Relative error tolerance.
            n_init (int): Initial number of samples. 
            n_limit (int): Maximum number of samples.
            error_fun (Union[str,callable]): Function mapping the approximate solution, absolute error tolerance, and relative error tolerance to the current error bound.

                - `'EITHER'`, the default, requires the approximation error must be below either the absolue *or* relative tolerance.
                    Equivalent to setting
                    ```python
                    error_fun = lambda sv,abs_tol,rel_tol: np.maximum(abs_tol,abs(sv)*rel_tol)
                    ```
                - `'BOTH'` requires the approximation error to be below both the absolue *and* relative tolerance. 
                    Equivalent to setting
                    ```python
                    error_fun = lambda sv,abs_tol,rel_tol: np.minimum(abs_tol,abs(sv)*rel_tol)
                    ```
            fudge (function): Positive function multiplying the finite sum of the Fourier coefficients specified in the cone of functions. 
            check_cone (bool): Whether or not to check if the function falls in the cone.
            ptransform (str): Periodization transform, see the options in [`AbstractIntegrand.f`][qmcpy.AbstractIntegrand.f].
        """
        super(CubQMCLatticeG,self).__init__(integrand,abs_tol,rel_tol,n_init,n_limit,fudge,check_cone,
            control_variates = [],
            control_variate_means = [],
            update_beta = False,
            ptransform = ptransform,
            ft = fftbr,
            omega = omega_fftbr,
            allowed_distribs = [Lattice],
            cast_complex = True,
            error_fun = error_fun)
        if self.discrete_distrib.order!='RADICAL INVERSE':
            raise ParameterError("CubLattice_g requires Lattice with 'RADICAL INVERSE' order")



from .abstract_stopping_criterion import AbstractStoppingCriterion
from ..util.data import Data

from ..discrete_distribution.abstract_discrete_distribution import AbstractDiscreteDistribution
from ..discrete_distribution.abstract_discrete_distribution import AbstractIIDDiscreteDistribution
from ..integrand import FinancialOption, Linear0, AbstractIntegrand
from ..true_measure import Gaussian, Uniform
from ..discrete_distribution import IIDStdUniform
from ..util import MaxSamplesWarning, ParameterError
import numpy as np
from scipy.optimize import root_scalar
from scipy.stats import norm
from time import time
import warnings


class CubMCG(AbstractStoppingCriterion):
    r"""
    IID Monte Carlo stopping criterion using Berry-Esseen inequalities in a two step method with guarantees for functions with bounded kurtosis.

    Examples:
        >>> ao = FinancialOption(IIDStdUniform(52,seed=7))
        >>> sc = CubMCG(ao,abs_tol=.05)
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        1.779
            bound_low       1.729
            bound_high      1.829
            bound_diff      0.100
            n_total         112314
            time_integrate  ...
        CubMCG (AbstractStoppingCriterion)
            abs_tol         0.050
            rel_tol         0
            n_init          2^(10)
            n_limit         2^(30)
            inflate         1.200
            alpha           0.010
            kurtmax         1.478
        FinancialOption (AbstractIntegrand)
            option          ASIAN
            call_put        CALL
            volatility      2^(-1)
            start_price     30
            strike_price    35
            interest_rate   0
            t_final         1
            asian_mean      ARITHMETIC
        BrownianMotion (AbstractTrueMeasure)
            time_vec        [0.019 0.038 0.058 ... 0.962 0.981 1.   ]
            drift           0
            mean            [0. 0. 0. ... 0. 0. 0.]
            covariance      [[0.019 0.019 0.019 ... 0.019 0.019 0.019]
                             [0.019 0.038 0.038 ... 0.038 0.038 0.038]
                             [0.019 0.038 0.058 ... 0.058 0.058 0.058]
                             ...
                             [0.019 0.038 0.058 ... 0.962 0.962 0.962]
                             [0.019 0.038 0.058 ... 0.962 0.981 0.981]
                             [0.019 0.038 0.058 ... 0.962 0.981 1.   ]]
            decomp_type     PCA
        IIDStdUniform (AbstractIIDDiscreteDistribution)
            d               52
            replications    1
            entropy         7

        Control variates 

        >>> iid = IIDStdUniform(52,seed=7)
        >>> ao = FinancialOption(iid,option="ASIAN")
        >>> eo = FinancialOption(iid,option="EUROPEAN")
        >>> lin0 = Linear0(iid)
        >>> sc = CubMCG(ao,abs_tol=.05,
        ...     control_variates = [eo,lin0],
        ...     control_variate_means = [eo.get_exact_value(),0])
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        1.787
            bound_low       1.737
            bound_high      1.837
            bound_diff      0.100
            n_total         52147
            time_integrate  ...
        CubMCG (AbstractStoppingCriterion)
            abs_tol         0.050
            rel_tol         0
            n_init          2^(10)
            n_limit         2^(30)
            inflate         1.200
            alpha           0.010
            kurtmax         1.478
            cv              [FinancialOption (AbstractIntegrand)
                                 option          EUROPEAN
                                 call_put        CALL
                                 volatility      2^(-1)
                                 start_price     30
                                 strike_price    35
                                 interest_rate   0
                                 t_final         1               Linear0 (AbstractIntegrand)]
            cv_mu           [4.211 0.   ]
        FinancialOption (AbstractIntegrand)
            option          ASIAN
            call_put        CALL
            volatility      2^(-1)
            start_price     30
            strike_price    35
            interest_rate   0
            t_final         1
            asian_mean      ARITHMETIC
        BrownianMotion (AbstractTrueMeasure)
            time_vec        [0.019 0.038 0.058 ... 0.962 0.981 1.   ]
            drift           0
            mean            [0. 0. 0. ... 0. 0. 0.]
            covariance      [[0.019 0.019 0.019 ... 0.019 0.019 0.019]
                             [0.019 0.038 0.038 ... 0.038 0.038 0.038]
                             [0.019 0.038 0.058 ... 0.058 0.058 0.058]
                             ...
                             [0.019 0.038 0.058 ... 0.962 0.962 0.962]
                             [0.019 0.038 0.058 ... 0.962 0.981 0.981]
                             [0.019 0.038 0.058 ... 0.962 0.981 1.   ]]
            decomp_type     PCA
        IIDStdUniform (AbstractIIDDiscreteDistribution)
            d               52
            replications    1
            entropy         7
        
        Relative tolerance 

        >>> ao = FinancialOption(IIDStdUniform(52,seed=7))
        >>> sc = CubMCG(ao,abs_tol=1e-3,rel_tol=5e-2)
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        1.743
            bound_low       1.661
            bound_high      1.825
            bound_diff      0.164
            n_total         27503
            time_integrate  ...
        CubMCG (AbstractStoppingCriterion)
            abs_tol         0.001
            rel_tol         0.050
            n_init          2^(10)
            n_limit         2^(30)
            inflate         1.200
            alpha           0.010
            kurtmax         1.478
        FinancialOption (AbstractIntegrand)
            option          ASIAN
            call_put        CALL
            volatility      2^(-1)
            start_price     30
            strike_price    35
            interest_rate   0
            t_final         1
            asian_mean      ARITHMETIC
        BrownianMotion (AbstractTrueMeasure)
            time_vec        [0.019 0.038 0.058 ... 0.962 0.981 1.   ]
            drift           0
            mean            [0. 0. 0. ... 0. 0. 0.]
            covariance      [[0.019 0.019 0.019 ... 0.019 0.019 0.019]
                             [0.019 0.038 0.038 ... 0.038 0.038 0.038]
                             [0.019 0.038 0.058 ... 0.058 0.058 0.058]
                             ...
                             [0.019 0.038 0.058 ... 0.962 0.962 0.962]
                             [0.019 0.038 0.058 ... 0.962 0.981 0.981]
                             [0.019 0.038 0.058 ... 0.962 0.981 1.   ]]
            decomp_type     PCA
        IIDStdUniform (AbstractIIDDiscreteDistribution)
            d               52
            replications    1
            entropy         7

        Relative tolerance and control variates 

        >>> iid = IIDStdUniform(52,seed=7)
        >>> ao = FinancialOption(iid,option="ASIAN")
        >>> eo = FinancialOption(iid,option="EUROPEAN")
        >>> lin0 = Linear0(iid)
        >>> sc = CubMCG(ao,abs_tol=1e-3,rel_tol=5e-2,
        ...     control_variates = [eo,lin0],
        ...     control_variate_means = [eo.get_exact_value(),0])
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        1.776
            bound_low       1.692
            bound_high      1.859
            bound_diff      0.167
            n_total         12074
            time_integrate  ...
        CubMCG (AbstractStoppingCriterion)
            abs_tol         0.001
            rel_tol         0.050
            n_init          2^(10)
            n_limit         2^(30)
            inflate         1.200
            alpha           0.010
            kurtmax         1.478
            cv              [FinancialOption (AbstractIntegrand)
                                 option          EUROPEAN
                                 call_put        CALL
                                 volatility      2^(-1)
                                 start_price     30
                                 strike_price    35
                                 interest_rate   0
                                 t_final         1               Linear0 (AbstractIntegrand)]
            cv_mu           [4.211 0.   ]
        FinancialOption (AbstractIntegrand)
            option          ASIAN
            call_put        CALL
            volatility      2^(-1)
            start_price     30
            strike_price    35
            interest_rate   0
            t_final         1
            asian_mean      ARITHMETIC
        BrownianMotion (AbstractTrueMeasure)
            time_vec        [0.019 0.038 0.058 ... 0.962 0.981 1.   ]
            drift           0
            mean            [0. 0. 0. ... 0. 0. 0.]
            covariance      [[0.019 0.019 0.019 ... 0.019 0.019 0.019]
                             [0.019 0.038 0.038 ... 0.038 0.038 0.038]
                             [0.019 0.038 0.058 ... 0.058 0.058 0.058]
                             ...
                             [0.019 0.038 0.058 ... 0.962 0.962 0.962]
                             [0.019 0.038 0.058 ... 0.962 0.981 0.981]
                             [0.019 0.038 0.058 ... 0.962 0.981 1.   ]]
            decomp_type     PCA
        IIDStdUniform (AbstractIIDDiscreteDistribution)
            d               52
            replications    1
            entropy         7
    
    **References:**

    1.  Fred J. Hickernell, Lan Jiang, Yuewei Liu, and Art B. Owen,  
        "Guaranteed conservative fixed width confidence intervals via Monte Carlo sampling,"  
        Monte Carlo and Quasi-Monte Carlo Methods 2012 (J. Dick, F. Y. Kuo, G. W. Peters, and I. H. Sloan, eds.), pp. 105-128,  
        Springer-Verlag, Berlin, 2014. DOI: 10.1007/978-3-642-41095-6_5

    2.  Sou-Cheng T. Choi, Yuhan Ding, Fred J. Hickernell, Lan Jiang, Lluis Antoni Jimenez Rugama,  
        Da Li, Jagadeeswaran Rathinavel, Xin Tong, Kan Zhang, Yizhi Zhang, and Xuan Zhou,  
        GAIL: Guaranteed Automatic Integration Library (Version 2.3) [MATLAB Software], 2019.  
        [http://gailgithub.github.io/GAIL_Dev/](http://gailgithub.github.io/GAIL_Dev/).  
        [https://github.com/GailGithub/GAIL_Dev/blob/master/Algorithms/IntegrationExpectation/meanMC_g.m](https://github.com/GailGithub/GAIL_Dev/blob/master/Algorithms/IntegrationExpectation/meanMC_g.m).
    """
    def __init__(self, 
                 integrand, 
                 abs_tol = 1e-2, 
                 rel_tol = 0., 
                 n_init = 1024, 
                 n_limit = 2**30,
                 inflate = 1.2, 
                 alpha = 0.01, 
                 control_variates = [], 
                 control_variate_means = [],
                 ):
        r"""
        Args:
            integrand (AbstractIntegrand): The integrand.
            abs_tol (np.ndarray): Absolute error tolerance.
            rel_tol (np.ndarray): Relative error tolerance.
            n_init (int): Initial number of samples. 
            n_limit (int): Maximum number of samples.
            inflate (float): Inflation factor $\geq 1$ to multiply by the variance estimate to make it more conservative.
            alpha (np.ndarray): Uncertainty level in $(0,1)$. 
            control_variates (list): Integrands to use as control variates, each with the same underlying discrete distribution instance.
            control_variate_means (np.ndarray): Means of each control variate. 
        """
        self.parameters = ['abs_tol','rel_tol','n_init','n_limit','inflate','alpha','kurtmax']
        # Set Attributes
        self.abs_tol = float(abs_tol)
        self.rel_tol = float(rel_tol)
        self.n_init = float(n_init)
        self.n_limit = float(n_limit)
        self.alpha = float(alpha)
        self.inflate = float(inflate)
        self.alpha_sigma = float(self.alpha) / 2.  # the uncertainty for variance estimation
        self.kurtmax = (n_init - 3) / (n_init - 1) + \
            (self.alpha_sigma * n_init) / (1 - self.alpha_sigma) * \
            (1 - 1. / self.inflate**2)**2
        # QMCPy Objs
        self.integrand = integrand
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib
        super(CubMCG,self).__init__(allowed_distribs=[AbstractIIDDiscreteDistribution],allow_vectorized_integrals=False)
        assert self.integrand.d_indv==()
        # control variates
        self.cv_mu = np.atleast_1d(control_variate_means)
        self.cv = control_variates
        if isinstance(self.cv,AbstractIntegrand):
            self.cv = [self.cv]
            self.cv_mu = self.cv_mu[None,...]
        assert isinstance(self.cv,list), "cv must be a list of AbstractIntegrand objects"
        for cv in self.cv:
            if (not isinstance(cv,AbstractIntegrand)) or (cv.discrete_distrib!=self.discrete_distrib) or (cv.d_indv!=self.integrand.d_indv):
                raise ParameterError('''
                        Each control variates discrete distribution must be an AbstractIntegrand instance 
                        with the same discrete distribution as the main integrand. d_indv must also match 
                        that of the main integrand instance for each control variate.''')
        self.ncv = len(self.cv)
        if self.ncv>0:
            assert self.cv_mu.shape==((self.ncv,)+self.integrand.d_indv), "Control variate means should have shape (len(control variates),d_indv)."
            self.parameters += ['cv','cv_mu']

    def integrate(self):
        t_start = time()
        data = Data(
            parameters = [
                'solution',
                'bound_low',
                'bound_high',
                'bound_diff',
                'n_total',
                'time_integrate'])
        data.xfull = np.empty((0,self.integrand.d))
        data.yfull = np.empty(0)
        if self.ncv>0:
            data.ycvfull = np.empty((self.ncv,0))
        x0 = self.discrete_distrib(n=self.n_init)
        data.xfull = np.concatenate([data.xfull,x0],0)
        y0 = self.integrand.f(x0)
        data.yfull = np.concatenate([data.yfull,y0],-1)
        if self.ncv>0:
            ycv0 = [None]*self.ncv
            for k in range(self.ncv):
                ycv0[k] = self.cv[k].f(x0)
            ycv0 = np.stack(ycv0,0)
            data.ycvfull = np.concatenate([data.ycvfull,ycv0],1)
            cvmuhats = ycv0.mean(-1)
            x4beta = ycv0-cvmuhats[:,None]
            y4beta = y0-y0.mean()
            self.beta = np.linalg.lstsq(x4beta.T,y4beta,rcond=None)[0]
            y0 = y0-((ycv0-self.cv_mu[:,None])*self.beta[:,None]).sum(0)
        data.sighat0 = y0.std(ddof=1)
        data.solution0 = y0.mean()
        sigma_up = self.inflate*data.sighat0
        if self.rel_tol == 0:
            self.alpha_mu = 1 - (1 - self.alpha) / (1 - self.alpha_sigma)
            toloversig = self.abs_tol / sigma_up
            n_mu,data.bound_half_width = self._nchebe(toloversig, self.alpha_mu, self.kurtmax, self.n_limit, sigma_up)
            data.n_mu = int(n_mu) 
            if (self.n_init+data.n_mu)>self.n_limit:
                # cannot generate this many new samples
                warning_s = """
                Already generated %d samples.
                Trying to generate %d new samples would exceed n_limit = %d.
                Will instead generate %d new samples to reach n_limit.""" \
                % (int(self.n_init),int(data.n_mu),int(self.n_limit),int(self.n_limit-self.n_init))
                warnings.warn(warning_s, MaxSamplesWarning)
                data.n_mu = self.n_limit-self.n_init
            x = self.discrete_distrib(n=data.n_mu)
            data.xfull = np.concatenate([data.xfull,x],0)
            y = self.integrand.f(x)
            data.yfull = np.concatenate([data.yfull,y],-1)
            if self.ncv>0:
                ycv = [None]*self.ncv
                for k in range(self.ncv):
                    ycv[k] = self.cv[k].f(x)
                ycv = np.stack(ycv,0)
                data.ycvfull = np.concatenate([data.ycvfull,ycv],1)
                y = y-((ycv-self.cv_mu[:,None])*self.beta[:,None]).sum(0)
            data.solution = y.mean()
            data.n_total = self.n_init+data.n_mu 
        else: # self.rel_tol > 0
            alphai = (self.alpha-self.alpha_sigma)/(2*(1-self.alpha_sigma)) # uncertainty to do iteration
            eps1 = self._ncbinv(self.n_init,alphai,self.kurtmax)
            data.bound_half_width = sigma_up*eps1
            data.tau = 1. # step of the iteration
            n = self.n_init # default initial sample size
            data.n_total = self.n_init
            while True:
                if (data.n_total + n) > self.n_limit:
                    # cannot generate this many new samples
                    warning_s = """
                    Already generated %d samples.
                    Trying to generate %d new samples would exceeds n_limit = %d.
                    Will instead generate %d samples to meet n_limit total samples. 
                    Note that error tolerances may no longer be satisfied""" \
                    % (int(data.n_total), int(n), int(self.n_limit), int(self.n_limit), int(self.n_limit-data.n_total))
                    warnings.warn(warning_s, MaxSamplesWarning)
                    n = self.n_limit-data.n_total
                x = self.discrete_distrib(n=n)
                data.xfull = np.concatenate([data.xfull,x],0)
                y = self.integrand.f(x)
                data.yfull = np.concatenate([data.yfull,y],-1)
                if self.ncv>0:
                    ycv = [None]*self.ncv
                    for k in range(self.ncv):
                        ycv[k] = self.cv[k].f(x)
                    ycv = np.stack(ycv,0)
                    data.ycvfull = np.concatenate([data.ycvfull,ycv],1)
                    y = y-((ycv-self.cv_mu[:,None])*self.beta[:,None]).sum(0)
                data.solution = y.mean()
                data.n_total += n
                if data.n_total>=self.n_limit: break
                lb_tol = _tol_fun(self.abs_tol, self.rel_tol, 0., data.solution-data.bound_half_width, 'max')
                ub_tol = _tol_fun(self.abs_tol, self.rel_tol, 0., data.solution+data.bound_half_width, 'max')
                delta_plus = (lb_tol + ub_tol) / 2.
                if delta_plus >= data.bound_half_width:
                    # stopping criterion met
                    delta_minus = (lb_tol - ub_tol) / 2.
                    data.solution += delta_minus # adjust solution a bit
                    break
                else:
                    candidate_tol = np.maximum(self.abs_tol,.95*self.rel_tol*abs(data.solution))
                    data.bound_half_width = np.minimum(data.bound_half_width/2.,candidate_tol)
                    data.tau += 1
                # update next uncertainty
                toloversig = data.bound_half_width / sigma_up
                alphai = 2**data.tau * (self.alpha - self.alpha_sigma) / (1 - self.alpha_sigma)
                n = int(self._nchebe(toloversig, alphai, self.kurtmax, self.n_limit, sigma_up)[0])
        data.bound_low = data.solution-data.bound_half_width
        data.bound_high = data.solution+data.bound_half_width
        data.bound_diff = data.bound_high-data.bound_low
        data.stopping_crit = self
        data.integrand = self.integrand
        data.true_measure = self.integrand.true_measure
        data.discrete_distrib = self.true_measure.discrete_distrib
        data.time_integrate = time()-t_start
        return data.solution,data

    def _nchebe(self, toloversig, alpha, kurtmax, n_budget, sigma_0_up):
        _b = -norm.ppf(np.finfo(float).eps)
        ncheb = np.ceil(1 / (alpha * toloversig**2))  # sample size by Chebyshev's Inequality
        A = 18.1139
        A1 = 0.3328
        A2 = 0.429  # three constants in Berry-Esseen inequality
        M3upper = kurtmax**(3. / 4)
        # the upper bound on the third moment by Jensen's inequality
        BEfun2 = lambda logsqrtn: \
            (norm.cdf(-np.exp(logsqrtn) * toloversig)
            + np.exp(-logsqrtn) * np.minimum(A1 * (M3upper + A2),
            A * M3upper / (1 + (np.exp(logsqrtn) * toloversig)**3))
            - alpha / 2.)
        # Berry-Esseen function, whose solution is the sample size needed
        logsqrtnCLT = np.log(norm.ppf(1 - alpha / 2) / toloversig)
        # sample size by CLT
        rsdata = root_scalar(BEfun2,x0=logsqrtnCLT,method='toms748',bracket=(-_b,_b))
        nbe = np.ceil(np.exp(2 * rsdata.root))
        # calculate Berry-Esseen n by fsolve function (scipy)
        ncb = np.minimum(np.minimum(ncheb, nbe), n_budget)  # take the min of two sample sizes
        logsqrtn = np.log(np.sqrt(ncb))
        BEfun3 = lambda toloversig: \
            (norm.cdf(-np.exp(logsqrtn) * toloversig)
            + np.exp(-logsqrtn) * np.minimum(A1 * (M3upper + A2),
            A * M3upper / (1 + (np.exp(logsqrtn) * toloversig)**3))
            - alpha / 2.)
        rsdata = root_scalar(BEfun3,x0=toloversig,method='toms748',bracket=(-_b,_b))
        err = rsdata.root * sigma_0_up
        return ncb, err

    def _ncbinv(self, n1, alpha1, kurtmax):
        _b = -norm.ppf(np.finfo(float).eps)
        NCheb_inv = 1/np.sqrt(n1*alpha1)
        # use Chebyshev inequality
        A = 18.1139
        A1 = 0.3328
        A2 = 0.429 # three constants in Berry-Esseen inequality
        M3upper = kurtmax**(3./4)
        # using Jensen's inequality to bound the third moment
        BEfun = lambda logsqrtb: \
            (norm.cdf(n1*logsqrtb) + np.minimum(A1*(M3upper+A2),
            A*M3upper/(1+(np.sqrt(n1)*logsqrtb)**3))/np.sqrt(n1)
            - alpha1/2)
        # Berry-Esseen inequality
        logsqrtb_clt = np.log(np.sqrt(norm.ppf(1-alpha1/2)/np.sqrt(n1)))
        # use CLT to get tolerance
        rsdata = root_scalar(BEfun,x0=logsqrtb_clt,method='toms748',bracket=(-_b,_b))
        NBE_inv = np.exp(2*rsdata.root)
        # use fsolve to get Berry-Esseen tolerance
        eps = np.minimum(NCheb_inv,NBE_inv)
        # take the min of Chebyshev and Berry Esseen tolerance
        return eps

    def set_tolerance(self, abs_tol=None, rel_tol=None, rmse_tol=None):
        assert rmse_tol is None, "rmse_tol not supported by this stopping criterion."
        if abs_tol != None: self.abs_tol = abs_tol
        if rel_tol != None: self.rel_tol = rel_tol

def _tol_fun(abs_tol, rel_tol, theta, mu, toltype):
    # """
    # Generalized error tolerance function.

    # Args:
    #     abs_tol (float): absolute error tolerance
    #     rel_tol (float): relative error tolerance
    #     theta (float): parameter in 'theta' case
    #     mu (float): true mean
    #     toltype (str): different options of tolerance function

    # Returns:
    #     float: tolerance as weighted sum of absolute and relative tolerance
    # """
    if toltype == 'combine':  # the linear combination of two tolerances
        # theta == 0 --> relative error tolerance
        # theta === 1 --> absolute error tolerance
        tol = theta * abs_tol + (1 - theta) * rel_tol * abs(mu)
    elif toltype == 'max':  # the max case
        tol = max(abs_tol, rel_tol * abs(mu))
    return tol



from ..util import _univ_repr

class Data(object):

    def __init__(self, parameters):
        self.parameters = parameters

    def __repr__(self):
        string = _univ_repr(self, 'Data', self.parameters + ['time_integrate'])
        if hasattr(self,"stopping_crit") and self.stopping_crit:
            string += '\n'+str(self.stopping_crit)
        if hasattr(self,"integrand") and self.integrand:
            string += '\n'+str(self.integrand)
        if hasattr(self,"true_measure") and self.true_measure:
            string += '\n'+str(self.true_measure)
        if hasattr(self,"discrete_distrib") and self.discrete_distrib:
            string += '\n'+str(self.discrete_distrib)
        return string



from .abstract_kernel import AbstractKernel
from .common_kernels import KernelGaussian
from ..util.transforms import tf_identity,tf_exp_eps,tf_exp_eps_inv,insert_batch_dims
import numpy as np 


class KernelMultiTask(AbstractKernel):

    r"""
    Multi-task kernel
    
    $$K((i,\boldsymbol{x}),(j,\boldsymbol{z})) = K_{\mathrm{task}}(i,j) K_{\mathrm{base}}(\boldsymbol{x},\boldsymbol{z})$$
    
    parameterized for $T$ tasks by a factor $\mathsf{F} \in \mathbb{R}^{T \times r}$ and a diagonal $\boldsymbol{v} \in \mathbb{R}^T$ so that 
    
    $$\left[K_{\mathrm{task}}(i,j)\right]_{i,j=1}^T = \mathsf{F} \mathsf{F}^T + \mathrm{diag}(\boldsymbol{v}).$$ 
    
    Examples:
        >>> kmt = KernelMultiTask(KernelGaussian(d=2),num_tasks=3,diag=[1,2,3])
        >>> x = np.random.rand(4,2) 
        >>> task = np.arange(3) 
        >>> kmt(task[1],task[1],x[0],x[0]).item()
        3.0
        >>> kmt(task,task,x[0],x[0])
        array([2., 3., 4.])
        >>> kmt(task[:,None],task[None,:],x[0],x[0]).shape
        (3, 3)
        >>> kmt(task[1],task[1],x,x,)
        array([3., 3., 3., 3.])
        >>> kmt(task,task,x[:3],x[:3])
        array([2., 3., 4.])
        >>> kmt(task[:,None],task[:,None],x,x).shape
        (3, 4)
        >>> v = kmt(task,task,x[:3,None,:],x[None,:3,:])
        >>> v.shape
        (3, 3)
        >>> np.allclose(v,kmt.base_kernel(x[:3,None,:],x[None,:3,:])*kmt.taskmat[task,task])
        True
        >>> kmt(task[:,None,None],task[:,None,None],x[:,None,:],x[None,:,:]).shape
        (3, 4, 4)
        >>> kmt(task[:,None,None,None],task[None,:,None,None],x[:,None,:],x[None,:,:]).shape
        (3, 3, 4, 4)

        Batched inference

        >>> kernel_base = KernelGaussian(d=10,shape_lengthscales=(5,1),shape_scale=(3,5,1))
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kmt = KernelMultiTask(kernel_base,
        ...     num_tasks=4,
        ...     shape_factor=(6,3,5,4,2),
        ...     diag = rng.uniform(low=0,high=1,size=(3,5,4)))
        >>> x = np.random.rand(8,10)
        >>> task = np.arange(4) 
        >>> kmt(task[0],task[0],x[0],x[0]).shape
        (6, 3, 5)
        >>> kmt(task,task,x[0],x[0]).shape
        (6, 3, 5, 4)
        >>> kmt(task[0],task[0],x,x).shape
        (6, 3, 5, 8)
        >>> v = kmt(task,task,x[:4,None,:],x[None,:4,:])
        >>> v.shape
        (6, 3, 5, 4, 4)
        >>> kmat_x = kmt.base_kernel(x[:4,None,:],x[None,:4,:])
        >>> kmat_x.shape
        (3, 5, 4, 4)
        >>> kmat_tasks = kmt.taskmat[...,task,task]
        >>> kmat_tasks.shape
        (6, 3, 5, 4)
        >>> np.allclose(v,kmat_tasks[...,None,:]*kmat_x)
        True
        >>> np.allclose(v,kmat_tasks[...,:,None]*kmat_x)
        False
        >>> kmt(task[:,None,None,None],task[None,:,None,None],x[:,None,:],x[None,:,:]).shape
        (6, 3, 5, 4, 4, 8, 8)
        >>> kmt(task[:,None],task[None,:],x[:,None,None,None,:],x[None,:,None,None,:]).shape
        (6, 3, 5, 8, 8, 4, 4)

        Integrals 

        >>> kernel = KernelGaussian(
        ...     d = 2,
        ...     scale = rng.uniform(low=0,high=1,size=(3,1)),
        ...     lengthscales = rng.uniform(low=0,high=1,size=(1,2)))
        >>> kmt = KernelMultiTask(
        ...     base_kernel=kernel,
        ...     num_tasks=5,
        ...     diag = rng.uniform(low=0,high=1,size=(6,3,5)),
        ...     factor = rng.uniform(low=0,high=1,size=(3,5,2)))
        >>> task = np.arange(5) 
        >>> kmt.double_integral_01d(task0=task[0],task1=task[1]).shape
        (6, 3)
        >>> kmt.double_integral_01d(task0=task,task1=task).shape
        (6, 3, 5)
        >>> kmt.double_integral_01d(task0=task[:,None],task1=task[None,:]).shape
        (6, 3, 5, 5)
        >>> x = rng.uniform(low=0,high=1,size=(7,4,2))
        >>> kmt.single_integral_01d(task0=task[0],task1=task[1],x=x).shape
        (6, 3, 7, 4)
        >>> kmt.single_integral_01d(task0=task[:,None,None],task1=task[:,None,None],x=x).shape
        (6, 3, 5, 7, 4)
        >>> kmt.single_integral_01d(task0=task[:,None,None,None],task1=task[None,:,None,None],x=x).shape
        (6, 3, 5, 5, 7, 4)

        Cholesky Construction 

        >>> kmt = KernelMultiTask(
        ...     KernelGaussian(5),
        ...     num_tasks = 3,
        ...     method = "CHOLESKY",
        ...     factor = 2,
        ...     diag = 1.5,
        ...     )
        >>> kmt.taskmat
        array([[ 2.25,  3.  ,  3.  ],
               [ 3.  ,  6.25,  7.  ],
               [ 3.  ,  7.  , 10.25]])
        >>> kmt = KernelMultiTask(
        ...     KernelGaussian(5),
        ...     num_tasks = 3,
        ...     method = "CHOLESKY",
        ...     factor = 2,
        ...     diag = 1.5,
        ...     shape_factor = [2,3*(1+3)//2-3],
        ...     shape_diag = [3],
        ...     )
        >>> kmt.taskmat
        array([[[ 2.25,  3.  ,  3.  ],
                [ 3.  ,  6.25,  7.  ],
                [ 3.  ,  7.  , 10.25]],
        <BLANKLINE>
               [[ 2.25,  3.  ,  3.  ],
                [ 3.  ,  6.25,  7.  ],
                [ 3.  ,  7.  , 10.25]]])
        >>> kmt = KernelMultiTask(
        ...     KernelGaussian(5),
        ...     num_tasks = 3,
        ...     method = "CHOLESKY",
        ...     factor = 2,
        ...     diag = 1.5,
        ...     shape_factor = [3*(1+3)//2-3],
        ...     shape_diag = [2,3],
        ...     )
        >>> kmt.taskmat
        array([[[ 2.25,  3.  ,  3.  ],
                [ 3.  ,  6.25,  7.  ],
                [ 3.  ,  7.  , 10.25]],
        <BLANKLINE>
               [[ 2.25,  3.  ,  3.  ],
                [ 3.  ,  6.25,  7.  ],
                [ 3.  ,  7.  , 10.25]]])
        >>> kmt = KernelMultiTask(
        ...     KernelGaussian(5),
        ...     num_tasks = 3,
        ...     method = "CHOLESKY",
        ...     factor = 2,
        ...     diag = 1.5,
        ...     shape_factor = [2,3*(1+3)//2-3],
        ...     shape_diag = [2,3],
        ...     )
        >>> kmt.taskmat
        array([[[ 2.25,  3.  ,  3.  ],
                [ 3.  ,  6.25,  7.  ],
                [ 3.  ,  7.  , 10.25]],
        <BLANKLINE>
               [[ 2.25,  3.  ,  3.  ],
                [ 3.  ,  6.25,  7.  ],
                [ 3.  ,  7.  , 10.25]]])

        PyTorch
        
        >>> import torch 
        >>> kmt = KernelMultiTask(KernelGaussian(d=2,torchify=True),num_tasks=3,diag=[1,2,3])
        >>> x = torch.from_numpy(np.random.rand(4,2))
        >>> task = np.arange(3) 
        >>> kmt(task[1],task[1],x[0],x[0]).item()
        3.0
        >>> kmt(task,task,x[0],x[0])
        tensor([2., 3., 4.], dtype=torch.float64, grad_fn=<SelectBackward0>)
        >>> kmt(task[:,None],task[None,:],x[0],x[0]).shape
        torch.Size([3, 3])
        >>> kmt(task[1],task[1],x,x)
        tensor([3., 3., 3., 3.], dtype=torch.float64, grad_fn=<SelectBackward0>)
        >>> kmt(task,task,x[:3],x[:3])
        tensor([2., 3., 4.], dtype=torch.float64, grad_fn=<SelectBackward0>)
        >>> kmt(task[:,None],task[:,None],x,x).shape
        torch.Size([3, 4])
        >>> v = kmt(task,task,x[:3,None,:],x[None,:3,:])
        >>> v.shape
        torch.Size([3, 3])
        >>> torch.allclose(v,kmt.base_kernel(x[:3,None,:],x[None,:3,:])*kmt.taskmat[task,task])
        True
        >>> kmt(task[:,None,None],task[:,None,None],x[:,None,:],x[None,:,:]).shape
        torch.Size([3, 4, 4])
        >>> kmt(task[:,None,None,None],task[None,:,None,None],x[:,None,:],x[None,:,:]).shape
        torch.Size([3, 3, 4, 4])

        Batched inference

        >>> kernel_base = KernelGaussian(d=10,shape_lengthscales=(5,1),shape_scale=(3,5,1),torchify=True)
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kmt = KernelMultiTask(kernel_base,
        ...     num_tasks=4,
        ...     shape_factor=(6,3,5,4,2),
        ...     diag = rng.uniform(low=0,high=1,size=(3,5,4)))
        >>> x = torch.from_numpy(np.random.rand(8,10))
        >>> task = torch.arange(4) 
        >>> kmt(task[0],task[0],x[0],x[0]).shape
        torch.Size([6, 3, 5])
        >>> kmt(task,task,x[0],x[0]).shape
        torch.Size([6, 3, 5, 4])
        >>> kmt(task[0],task[0],x,x).shape
        torch.Size([6, 3, 5, 8])
        >>> v = kmt(task,task,x[:4,None,:],x[None,:4,:])
        >>> v.shape
        torch.Size([6, 3, 5, 4, 4])
        >>> kmat_x = kmt.base_kernel(x[:4,None,:],x[None,:4,:])
        >>> kmat_x.shape
        torch.Size([3, 5, 4, 4])
        >>> kmat_tasks = kmt.taskmat[...,task,task]
        >>> kmat_tasks.shape
        torch.Size([6, 3, 5, 4])
        >>> torch.allclose(v,kmat_tasks[...,None,:]*kmat_x)
        True
        >>> torch.allclose(v,kmat_tasks[...,:,None]*kmat_x)
        False
        >>> kmt(task[:,None,None,None],task[None,:,None,None],x[:,None,:],x[None,:,:]).shape
        torch.Size([6, 3, 5, 4, 4, 8, 8])
        >>> kmt(task[:,None],task[None,:],x[:,None,None,None,:],x[None,:,None,None,:]).shape
        torch.Size([6, 3, 5, 8, 8, 4, 4])
        >>> kmt.factor.dtype
        torch.float32
        >>> kmt.diag.dtype
        torch.float64

        Integrals 
        
        >>> kernel = KernelGaussian(
        ...     d = 2,
        ...     torchify = True,
        ...     scale = rng.uniform(low=0,high=1,size=(3,1)),
        ...     lengthscales = rng.uniform(low=0,high=1,size=(1,2)))
        >>> kmt = KernelMultiTask(
        ...     base_kernel=kernel,
        ...     num_tasks=5,
        ...     diag = rng.uniform(low=0,high=1,size=(6,3,5)),
        ...     factor = rng.uniform(low=0,high=1,size=(3,5,2)))
        >>> task = torch.arange(5) 
        >>> kmt.double_integral_01d(task0=task[0],task1=task[1]).shape
        torch.Size([6, 3])
        >>> kmt.double_integral_01d(task0=task,task1=task).shape
        torch.Size([6, 3, 5])
        >>> kmt.double_integral_01d(task0=task[:,None],task1=task[None,:]).shape
        torch.Size([6, 3, 5, 5])
        >>> x = torch.from_numpy(rng.uniform(low=0,high=1,size=(7,4,2)))
        >>> kmt.single_integral_01d(task0=task[0],task1=task[1],x=x).shape
        torch.Size([6, 3, 7, 4])
        >>> kmt.single_integral_01d(task0=task[:,None,None],task1=task[:,None,None],x=x).shape
        torch.Size([6, 3, 5, 7, 4])
        >>> kmt.single_integral_01d(task0=task[:,None,None,None],task1=task[None,:,None,None],x=x).shape
        torch.Size([6, 3, 5, 5, 7, 4])

        Cholesky Construction 

        >>> kmt = KernelMultiTask(
        ...     KernelGaussian(5,torchify=True),
        ...     num_tasks = 3,
        ...     method = "CHOLESKY",
        ...     factor = 2,
        ...     diag = 1.5,
        ...     )
        >>> kmt.taskmat
        tensor([[ 2.2500,  3.0000,  3.0000],
                [ 3.0000,  6.2500,  7.0000],
                [ 3.0000,  7.0000, 10.2500]], grad_fn=<ViewBackward0>)
        >>> kmt = KernelMultiTask(
        ...     KernelGaussian(5,torchify=True),
        ...     num_tasks = 3,
        ...     method = "CHOLESKY",
        ...     factor = 2,
        ...     diag = 1.5,
        ...     shape_factor = [2,3*(1+3)//2-3],
        ...     shape_diag = [3],
        ...     )
        >>> kmt.taskmat
        tensor([[[ 2.2500,  3.0000,  3.0000],
                 [ 3.0000,  6.2500,  7.0000],
                 [ 3.0000,  7.0000, 10.2500]],
        <BLANKLINE>
                [[ 2.2500,  3.0000,  3.0000],
                 [ 3.0000,  6.2500,  7.0000],
                 [ 3.0000,  7.0000, 10.2500]]], grad_fn=<ViewBackward0>)
        >>> kmt = KernelMultiTask(
        ...     KernelGaussian(5,torchify=True),
        ...     num_tasks = 3,
        ...     method = "CHOLESKY",
        ...     factor = 2,
        ...     diag = 1.5,
        ...     shape_factor = [3*(1+3)//2-3],
        ...     shape_diag = [2,3],
        ...     )
        >>> kmt.taskmat
        tensor([[[ 2.2500,  3.0000,  3.0000],
                 [ 3.0000,  6.2500,  7.0000],
                 [ 3.0000,  7.0000, 10.2500]],
        <BLANKLINE>
                [[ 2.2500,  3.0000,  3.0000],
                 [ 3.0000,  6.2500,  7.0000],
                 [ 3.0000,  7.0000, 10.2500]]], grad_fn=<ViewBackward0>)
        >>> kmt = KernelMultiTask(
        ...     KernelGaussian(5,torchify=True),
        ...     num_tasks = 3,
        ...     method = "CHOLESKY",
        ...     factor = 2,
        ...     diag = 1.5,
        ...     shape_factor = [2,3*(1+3)//2-3],
        ...     shape_diag = [2,3],
        ...     )
        >>> kmt.taskmat
        tensor([[[ 2.2500,  3.0000,  3.0000],
                 [ 3.0000,  6.2500,  7.0000],
                 [ 3.0000,  7.0000, 10.2500]],
        <BLANKLINE>
                [[ 2.2500,  3.0000,  3.0000],
                 [ 3.0000,  6.2500,  7.0000],
                 [ 3.0000,  7.0000, 10.2500]]], grad_fn=<ViewBackward0>)
    """

    def __init__(self,
            base_kernel,
            num_tasks, 
            factor = 1.,
            diag =  1.,
            shape_factor = None, 
            shape_diag = None,
            tfs_factor = (tf_identity,tf_identity),
            tfs_diag = (tf_exp_eps_inv,tf_exp_eps),
            requires_grad_factor = True, 
            requires_grad_diag = True,
            rank_factor = 1,
            method = "LOW RANK"
            ):
        r"""
        Args:
            base_kernel (AbstractKernel): $K_{\mathrm{base}}$. 
            num_tasks (int): Number of tasks $T>1$. 
            factor (Union[np.ndarray,torch.Tensor]): Factor $\mathsf{F}$. 
            diag (Union[np.ndarray,torch.Tensor]): Diagonal parameter $\boldsymbol{v}$. 
            shape_factor (list): Shape of `factor` when `np.isscalar(factor)`. 
            shape_diag (list): Shape of `diag` when `np.isscalar(diag)`. 
            tfs_factor (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            tfs_diag (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            requires_grad_factor (bool): If `True` and `torchify`, set `requires_grad=True` for `factor`.
            requires_grad_diag (bool): If `True` and `torchify`, set `requires_grad=True` for `diag`.
            method (str): `"LOW RANK"` or "CHOLESKY"
        """
        assert isinstance(base_kernel,AbstractKernel)
        super().__init__(
            d = base_kernel.d,
            torchify = base_kernel.torchify,
            device = base_kernel.device,
            compile_call = False,
            comiple_call_kwargs = {})
        self.base_kernel = base_kernel
        self.AUTOGRADKERNEL = base_kernel.AUTOGRADKERNEL 
        assert np.isscalar(num_tasks) and num_tasks%1==0
        self.num_tasks = num_tasks
        assert np.isscalar(rank_factor) and rank_factor%1==0 and 0<=rank_factor<=self.num_tasks
        self.method = str(method).upper().replace("_"," ").strip()
        if self.method=="LOW RANK":
            if shape_factor is None:
                shape_factor = [self.num_tasks,rank_factor]
            factor_endsize_ops = list(range(self.num_tasks+1))
            diag_endsize_ops = [1,self.num_tasks]
        elif self.method=="CHOLESKY":
            if self.torchify:
                self.lti0,self.lti1 = self.npt.tril_indices(num_tasks,num_tasks,-1)
            else:
                self.lti0,self.lti1 = self.npt.tril_indices(num_tasks,-1)
            if shape_factor is None:
                shape_factor = [len(self.lti0)]
            factor_endsize_ops = [len(self.lti0)]
            diag_endsize_ops = [self.num_tasks]
        else:
            raise Exception("invalid method = %s, must be in ['LOW RANK','CHOLESKY']"%self.method)
        self.raw_factor = self.parse_assign_param(
            pname = "factor",
            param = factor,
            shape_param = shape_factor,
            requires_grad_param = requires_grad_factor,
            tfs_param = tfs_factor,
            endsize_ops = factor_endsize_ops,
            constraints = [])
        self.tfs_factor = tfs_factor
        if self.method=="LOW RANK":
            assert self.raw_factor.shape[-2]==self.num_tasks
        self.raw_diag = self.parse_assign_param(
            pname = "diag",
            param = diag, 
            shape_param = [self.num_tasks] if shape_diag is None else shape_diag,
            requires_grad_param = requires_grad_diag,
            tfs_param = tfs_diag,
            endsize_ops = diag_endsize_ops,
            constraints = ["NON-NEGATIVE"])
        self.tfs_diag = tfs_diag
        self.eye_num_tasks = self.npt.eye(self.num_tasks,**self.nptkwargs)
        self.batch_param_names.append("taskmat")
        self._nbdim_base = None
    
    @property
    def nbdim_base(self):
        if self._nbdim_base is None:
            self._nbdim_base = self.base_kernel.nbdim
        return self._nbdim_base
    
    @property
    def factor(self):
        return self.tfs_factor[1](self.raw_factor)
    
    @property
    def diag(self):
        return self.tfs_diag[1](self.raw_diag)

    @property
    def taskmat(self):
        factor = self.factor
        diag = self.diag
        if self.method=="LOW RANK":
            taskmat = self.npt.einsum("...ij,...kj->...ik",factor,factor)+diag[...,None]*self.eye_num_tasks
        elif self.method=="CHOLESKY":
            L = diag[...,None]*self.eye_num_tasks+self.npt.zeros(list(self.factor.shape[:-1])+[1,1],**self.nptkwargs)
            L[...,self.lti0,self.lti1] = self.factor
            taskmat = self.npt.einsum("...ij,...kj->...ik",L,L)
        else:
            raise Exception("invalid method = %s, must be in ['LOW RANK','CHOLESKY']"%self.method)
        return taskmat
    
    def _parsed__call__(self, task0, task1, k_x):
        k_x = k_x[...,None]
        nnbdim_v = k_x.ndim-self.nbdim_base-1
        kmat_tasks = self.taskmat
        nbdim = kmat_tasks.ndim-2
        kmat_tasks = kmat_tasks[...,task0,task1,None]
        nnbdim_kmat = kmat_tasks.ndim-nbdim-1
        commondim = max(nnbdim_v,nnbdim_kmat)
        k_x = insert_batch_dims(k_x,commondim-nnbdim_v,self.nbdim_base)
        kmat_tasks = insert_batch_dims(kmat_tasks,commondim-nnbdim_kmat,nbdim)
        kmat = k_x*kmat_tasks 
        return kmat[...,0]
    
    def __call__(self, task0, task1, x0, x1, beta0=None, beta1=None, c=None):
        r"""
        Evaluate the kernel with (optional) partial derivatives 

        $$\sum_{\ell=1}^p c_\ell \partial_{\boldsymbol{x}_0}^{\boldsymbol{\beta}_{\ell,0}} \partial_{\boldsymbol{x}_1}^{\boldsymbol{\beta}_{\ell,1}} K((i_0,\boldsymbol{x}_0),(i_1,\boldsymbol{x}_1)).$$
        
        Args:
            task0 (Union[int,np.ndarray,torch.Tensor]): First task indices $i_0$. 
            task1 (Union[int,np.ndarray,torch.Tensor]): Second task indices $i_1$.
            x0 (Union[np.ndarray,torch.Tensor]): Shape `x0.shape=(...,d)` first input to kernel.
            x1 (Union[np.ndarray,torch.Tensor]): Shape `x1.shape=(...,d)` second input to kernel. 
            beta0 (Union[np.ndarray,torch.Tensor]): Shape `beta0.shape=(p,d)` derivative orders with respect to first inputs, $\boldsymbol{\beta}_0$.
            beta1 (Union[np.ndarray,torch.Tensor]): Shape `beta1.shape=(p,d)` derivative orders with respect to first inputs, $\boldsymbol{\beta}_1$.
            c (Union[np.ndarray,torch.Tensor]): Shape `c.shape=(p,)` coefficients of derivatives.

        Returns:
            k (Union[np.ndarray,torch.Tensor]): Kernel evaluations with batched shape, see the doctests for examples. 
        """
        kmat_x = self.base_kernel.__call__(x0,x1,beta0,beta1,c)
        return self._parsed__call__(task0,task1,kmat_x)
    
    def single_integral_01d(self, task0, task1, x):
        r"""
        Evaluate the integral of the kernel over the unit cube

        $$\tilde{K}((i_0,\boldsymbol{x}),i_1) = \int_{[0,1]^d} K((i_0,\boldsymbol{x}),(i_1,\boldsymbol{z}) \; \mathrm{d} \boldsymbol{z}.$$
        
        Args:
            task0 (Union[int,np.ndarray,torch.Tensor]): First task indices $i_0$. 
            task1 (Union[int,np.ndarray,torch.Tensor]): Second task indices $i_1$.
            x (Union[np.ndarray,torch.Tensor]): Shape `x0.shape=(...,d)` first input to kernel with 
        
        Returns:
            tildek (Union[np.ndarray,torch.Tensor]): Shape `y.shape=x.shape[:-1]` integral kernel evaluations. 
        """
        kint_x = self.base_kernel.single_integral_01d(x)
        return self._parsed__call__(task0,task1,kint_x)
        
    def double_integral_01d(self, task0, task1):
        r"""
        Evaluate the integral of the kernel over the unit cube

        $$\tilde{K}(i_0,i_1) = \int_{[0,1]^d} \int_{[0,1]^d} K((i_0,\boldsymbol{x}),(i_1,\boldsymbol{z})) \; \mathrm{d} \boldsymbol{x} \; \mathrm{d} \boldsymbol{z}.$$
        
        Args:
            task0 (Union[int,np.ndarray,torch.Tensor]): First task indices $i_0$. 
            task1 (Union[int,np.ndarray,torch.Tensor]): Second task indices $i_1$.
        
        Returns:
            tildek (Union[np.ndarray,torch.Tensor]): Double integral kernel evaluations.
        """
        kint_x = self.base_kernel.double_integral_01d()
        return self._parsed__call__(task0,task1,kint_x)

class KernelMultiTaskDerivs(KernelMultiTask):
    def __init__(self,
            base_kernel,
            num_tasks, 
            ):
        super().__init__(
            base_kernel = base_kernel, 
            num_tasks = num_tasks, 
            factor = 1.,
            diag =  0.,
            requires_grad_factor = False, 
            requires_grad_diag = False,
            tfs_factor = (tf_identity,tf_identity),
            tfs_diag = (tf_identity,tf_identity),
            rank_factor = 1,
        )
        




from .abstract_stopping_criterion import AbstractStoppingCriterion
from .abstract_cub_mlmc import AbstractCubMLMC
from ..util.data import Data

from ..discrete_distribution import IIDStdUniform
from ..discrete_distribution.abstract_discrete_distribution import AbstractIIDDiscreteDistribution
from ..true_measure import Gaussian
from ..integrand import FinancialOption
from ..util import MaxSamplesWarning, ParameterError, MaxLevelsWarning, ParameterWarning
import numpy as np
from scipy.stats import norm
from time import time
import warnings


class CubMLMCCont(AbstractCubMLMC):
    r"""
    Multilevel IID Monte Carlo stopping criterion with continuation.
    
    Examples:
        >>> fo = FinancialOption(IIDStdUniform(seed=7))
        >>> sc = CubMLMCCont(fo,abs_tol=1.5e-2)
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        1.771
            n_total         2291120
            levels          3
            n_level         [1094715  222428   79666     912     256]
            mean_level      [1.71  0.048 0.012]
            var_level       [21.826  1.768  0.453]
            cost_per_sample [2. 4. 8.]
            alpha           1.970
            beta            1.965
            gamma           1.000
            time_integrate  ...
        CubMLMCCont (AbstractStoppingCriterion)
            rmse_tol        0.006
            n_init          2^(8)
            levels_min      2^(1)
            levels_max      10
            n_tols          10
            inflate         1.668
            theta_init      2^(-1)
            theta           0.010
        FinancialOption (AbstractIntegrand)
            option          ASIAN
            call_put        CALL
            volatility      2^(-1)
            start_price     30
            strike_price    35
            interest_rate   0
            t_final         1
            asian_mean      ARITHMETIC
        BrownianMotion (AbstractTrueMeasure)
            time_vec        1
            drift           0
            mean            0
            covariance      1
            decomp_type     PCA
        IIDStdUniform (AbstractIIDDiscreteDistribution)
            d               1
            replications    1
            entropy         7

    **References:**

    1. [https://github.com/PieterjanRobbe/MultilevelEstimators.jl](https://github.com/PieterjanRobbe/MultilevelEstimators.jl).
    """

    def __init__(self, 
                 integrand, 
                 abs_tol = .05, 
                 rmse_tol = None, 
                 n_init = 256, 
                 n_limit = 1e10,
                 inflate = 100**(1/9),
                 alpha = .01, 
                 levels_min = 2,
                 levels_max = 10, 
                 n_tols = 10, 
                 theta_init = 0.5,
                 ):
        r"""
        Args:
            integrand (AbstractIntegrand): The integrand.
            abs_tol (np.ndarray): Absolute error tolerance.
            rmse_tol (np.ndarray): Root mean squared error tolerance. 
                If supplied, then absolute tolerance and alpha are ignored in favor of the rmse tolerance. 
            n_init (int): Initial number of samples. 
            n_limit (int): Maximum number of samples.
            inflate (float): Coarser tolerance multiplication factor $\geq 1$.
            alpha (np.ndarray): Uncertainty level in $(0,1)$. 
            levels_min (int): Minimum level of refinement $\geq 2$.
            levels_max (int): Maximum level of refinement $\geq$ `levels_min`.
            n_tols (int): Number of coarser tolerances to run.
            theta_init (float): Initial error splitting constant.
        """
        self.parameters = ['rmse_tol','n_init','levels_min','levels_max','n_tols',
            'inflate','theta_init','theta']
        if levels_min < 2:
            raise ParameterError('needs levels_min >= 2')
        if levels_max < levels_min:
            raise ParameterError('needs levels_max >= levels_min')
        if n_init <= 0:
            raise ParameterError('needs n_init > 0')
        # initialization
        if rmse_tol:
            self.target_tol = float(rmse_tol)
        else: # use absolute tolerance
            self.target_tol =  float(abs_tol) / norm.ppf(1-alpha/2)
        self.n_init = n_init
        self.n_limit = n_limit
        self.levels_min = levels_min
        self.levels_max = levels_max
        self.theta_init = theta_init
        self.theta = theta_init
        self.n_tols = n_tols
        self.inflate = inflate
        self.integrand = integrand
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib
        self.alpha0 = -1 
        self.beta0 = -1 
        self.gamma0 = -1
        self.alpha = alpha
        self.inflate = inflate
        assert self.inflate>=1
        assert 0<self.alpha<1
        super(CubMLMCCont,self).__init__(allowed_distribs=[AbstractIIDDiscreteDistribution],allow_vectorized_integrals=False)

    def integrate(self):
        t_start = time()
        data = self._construct_data()
        # Loop over coarser tolerances
        for t in range(self.n_tols):
            self.rmse_tol = self.inflate**(self.n_tols-t-1)*self.target_tol # Set new target tolerance
            self._integrate(data)
        data.stopping_crit = self
        data.integrand = self.integrand
        data.true_measure = self.integrand.true_measure
        data.discrete_distrib = self.true_measure.discrete_distrib
        data.time_integrate = time()-t_start
        return data.solution,data
    
    def _construct_data(self):
        data = Data(parameters=[
            'solution',
            'n_total',
            'levels',
            'n_level',
            'mean_level',
            'var_level', 
            'cost_per_sample',
            'alpha',
            'beta',
            'gamma'])
        data.levels = int(self.levels_min)
        data.n_level = np.zeros(data.levels+1,dtype=int)
        data.sum_level = np.zeros((2,data.levels+1))
        data.cost_level = np.zeros(data.levels+1)
        data.diff_n_level = self.n_init*np.ones(data.levels+1,dtype=int)
        data.alpha = np.maximum(0,self.alpha0)
        data.beta = np.maximum(0,self.beta0)
        data.gamma = np.maximum(0,self.gamma0)
        data.level_integrands = []
        return data

    def _integrate(self, data):
        self.theta = self.theta_init
        data.levels = int(self.levels_min)
        self._update_data(data) # Take warm-up samples if none have been taken so far

        converged = False
        while not converged:

            # Check if we already have samples at the finest level
            if not data.n_level[data.levels] > 0:
                # This takes n_init warm-up samples at the finest level
                data.diff_n_level = np.hstack((data.diff_n_level,self.n_init))
                self._update_data(data)
                # Alternatively, evaluate optimal number of samples and take between 2 and n_init samples
                #data.diff_n_level = self._get_next_samples(data)
                #data.diff_n_level[:data.levels] = 0
                #data.diff_n_level[data.levels] = max(3, min(self.n_init, data.diff_n_level[data.levels]))

            # Update splitting parameter
            self._update_theta(data)

            # Set optimal number of additional samples
            n_samples = self._get_next_samples(data)
            data.diff_n_level = np.maximum(0, n_samples-data.n_level[:data.levels+1])
            
            # Check if over sample budget
            if (data.n_total + data.diff_n_level.sum()) > self.n_limit:
                warning_s = """
                Already generated %d samples.
                Trying to generate %d new samples, which would exceed n_limit = %d.
                Stopping integration process.
                Note that error tolerances may no longer be satisfied""" \
                % (int(data.n_total), int(data.diff_n_level.sum()), int(self.n_limit))
                warnings.warn(warning_s, MaxSamplesWarning)
                break

            # Take additional samples
            self._update_data(data)
            data.n_total += data.diff_n_level.sum()

            # Check for convergence
            converged = self._rmse(data) < self.rmse_tol
            if not converged:
                if data.levels == self.levels_max:
                    warnings.warn(
                        'Failed to achieve weak convergence. levels == levels_max.',
                        MaxLevelsWarning)
                    converged = True
                else:
                    self._add_level(data)

        data.diff_n_level.fill(0)
        data.solution = (data.sum_level[0,:data.levels+1]/data.n_level[:data.levels+1]).sum()
        data.levels += 1

    def _update_theta(self, data):
        # Update error splitting parameter
        self.theta = max(0.01, min(0.5, (self._bias(data,len(data.n_level)-1)/self.rmse_tol)**2))

    def _rmse(self, data):
        # Returns an estimate for the root mean square error
        return np.sqrt(self._mse(data))

    def _mse(self, data):
        # Returns an estimate for the mean square error
        return (1-self.theta)*self._varest(data) + self.theta*self._bias(data,data.levels)**2

    def _varest(self, data):
        # Returns the variance of the estimator
        return (data.var_level/data.n_level[:data.levels+1]).sum()

    def _bias(self, data, level):
        # Returns an estimate for the bias
        mean_level = data.sum_level[0, :]/data.n_level
        A = np.ones((2,2))
        A[:,0] = range(level-1, level+1)
        y = np.log2(abs(mean_level[level-1:level+1]))
        x = np.linalg.lstsq(A, y, rcond=None)[0]
        alpha = max(.5,-x[0])
        return 2**(x[1]+(level+1)*x[0]) / (2**alpha - 1)



import numpy as np 
import os
from typing import Union

def plot_proj(
        sampler, 
        n = 64,
        d_horizontal = 1, 
        d_vertical = 2,
        math_ind = True, 
        marker_size = 5, 
        figfac = 5, 
        fig_title = 'Projection of Samples', 
        axis_pad = 0, 
        want_grid = True, 
        font_family = "sans-serif", 
        where_title = 1, 
        **kwargs):
    """
    Args:
        sampler (DiscreteDistribution,TrueMeasure): The generator of samples to be plotted.
        n (Union[int,list]): The number of samples or a list of samples(used for extensibility) to be plotted.
        d_horizontal (Union[int,list]): The dimension or list of dimensions to be plotted on the horizontal axes. 
        d_vertical (Union[int,list]): The dimension or list of dimensions to be plotted on the vertical axes. 
        math_ind (bool): Setting to `True` will enable user to pass in math indices.
        marker_size (float): The marker size (typographic points are 1/72 in.).
        figfac (float): The figure size factor.
        fig_title (str): The title of the figure.
        axis_pad (float): The padding of the axis so that points on the boundaries can be seen.
        want_grid (bool): Setting to `True` will enable grid on the plot.
        font_family (str): The font family of the plot.
        where_title (float): the position of the title on the plot. Default value is 1.
        **kwargs (dict): Additional keyword arguments passed to `matplotlib.pyplot.scatter`.
    """
    try:
        import matplotlib.pyplot as plt
        from matplotlib import colors
        dir_path = os.path.dirname(os.path.realpath(__file__))
        plt.style.use(os.path.join(dir_path, "qmcpy.mplstyle"))
    except:
        raise ImportError("Missing matplotlib.pyplot as plt, Matplotlib must be installed to run plot_proj function")
    plt.rcParams['font.family'] = font_family 
    n = np.atleast_1d(n)
    d_horizontal = np.atleast_1d(d_horizontal)
    d_vertical = np.atleast_1d(d_vertical)
    samples = sampler(n[n.size - 1])    
    d = samples.shape[1]
    fig, ax = plt.subplots(nrows=d_horizontal.size, ncols=d_vertical.size, figsize=(figfac*d_horizontal.size, figfac*d_vertical.size),squeeze=False)                    
    
    colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
    for i in range(d_horizontal.size):         
        for j in range(d_vertical.size):
                n_min = 0
                for m in range(n.size):
                    n_max = n[m]  
                    if(d_horizontal[i] == d_vertical[j]):
                        ax[i,j].remove()  
                        break
                    if(math_ind is True):
                        x = d_horizontal[i] - 1
                        y = d_vertical[j] - 1
                        x_label_num = d_horizontal[i]
                        y_label_num = d_vertical[j]
                    else:
                        x = d_horizontal[i]
                        y = d_vertical[j]
                        x_label_num = d_horizontal[i] + 1
                        y_label_num = d_vertical[j] + 1
                    
                    if type(sampler).__name__=="AbstractDiscreteDistribution":
                        ax[i,j].set_xlim([0-axis_pad,1+axis_pad])
                        ax[i,j].set_ylim([0-axis_pad,1+axis_pad])
                        ax[i,j].set_xticks([0,1/4,1/2,3/4,1])
                        ax[i,j].set_yticks([0,1/4,1/2,3/4,1])
                        ax[i,j].set_aspect(1)
                        if not want_grid:
                            ax[i,j].grid(False) 
                            ax[i,j].tick_params(axis='both', which='both', direction='in', length=5)
                        ax[i,j].grid(want_grid)
                        x_label = r'$x_{i%d}$'%(x_label_num)
                        y_label = r'$x_{i%d}$'%(y_label_num)
                    else:
                        x_label = r'$t_{i%d}$'%(x_label_num)
                        y_label = r'$t_{i%d}$'%(y_label_num)    
                    
                    ax[i,j].set_xlabel(x_label,fontsize = 15)
                    if(d > 1):
                        ax[i,j].set_ylabel(y_label,fontsize = 15)
                        y_axis = samples[n_min:n_max,y]
                    else:
                        y_axis = []
                        for h in range(n_max-n_min):
                            y_axis.append(0.5)
                    ax[i,j].scatter(samples[n_min:n_max,x],y_axis,s=marker_size,color=colors[m],label='n_min = %d, n_max = %d'%(n_min,n_max),**kwargs)
                    n_min = n[m]
    plt.suptitle(fig_title,fontsize = 20, y = where_title)
    #fig.text(0.55,0.55,fig_title, ha = 'center', va = 'center', fontsize = 20) %replaced by plt.suptitle
    fig.tight_layout()#pad=2)
    return fig, ax